from sqlalchemy.orm import Query

from cg.models.orders.constants import OrderType
from cg.store.filters.status_ordertype_application_filters import filter_applications_by_order_type
from cg.store.models import Application, OrderTypeApplication
from cg.store.store import Store
from tests.store_helpers import StoreHelpers


def test_filter_applications_by_order_type(applications_store: Store, helpers: StoreHelpers):
    # GIVEN a store with applications
    applications: list[Application] = applications_store.get_applications()
    assert applications

    # GIVEN an order type
    order_type = OrderType.PACBIO_LONG_READ

    # GIVEN that one application has the given order type
    helpers.add_application_order_type(
        store=applications_store, application=applications[0], order_types=[order_type]
    )

    # GIVEN that another application has a different order type
    helpers.add_application_order_type(
        store=applications_store, application=applications[1], order_types=[OrderType.BALSAMIC]
    )

    # WHEN filtering applications by order type
    order_type_applications: Query = applications_store._get_query(table=OrderTypeApplication)
    filtered_order_type_applications: Query = filter_applications_by_order_type(
        order_type_applications=order_type_applications, order_type=order_type
    )

    # THEN assert that only the applications with the given order type are returned
    assert filtered_order_type_applications.count() == 1
    assert filtered_order_type_applications.first().order_type == order_type



from datetime import datetime

from cg.server.dto.samples.samples_response import SamplesResponse
from cg.services.sample_service.dto_mappers import create_sample_dto
from cg.store.models import Sample


def get_cancel_comment(user_name: str) -> str:
    date: str = datetime.now().strftime("%Y-%m-%d")
    return f"Cancelled {date} by {user_name}"


def get_confirmation_message(sample_ids: list[str], case_ids: list[str]) -> str:
    message = f"Cancelled {len(sample_ids)} samples. "
    if case_ids:
        cases = ", ".join(case_ids)
        message += f"Found {len(case_ids)} cases with additional samples: {cases}."
    else:
        message += "No case contained additional samples."
    return message


def create_samples_response(samples: list[Sample]) -> SamplesResponse:
    sample_dtos = []
    for sample in samples:
        sample_dtos.append(create_sample_dto(sample))
    return SamplesResponse(samples=sample_dtos, total=len(samples))



import pytest

from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.services.deliver_files.tag_fetcher.bam_service import (
    BamDeliveryTagsFetcher,
)
from cg.services.fastq_concatenation_service.fastq_concatenation_service import (
    FastqConcatenationService,
)
from cg.services.deliver_files.tag_fetcher.sample_and_case_service import (
    SampleAndCaseDeliveryTagsFetcher,
)
from cg.services.deliver_files.file_fetcher.analysis_service import (
    AnalysisDeliveryFileFetcher,
)
from cg.services.deliver_files.file_fetcher.raw_data_service import (
    RawDataDeliveryFileFetcher,
)
from cg.services.deliver_files.file_formatter.service import (
    DeliveryFileFormatter,
)
from cg.services.deliver_files.file_formatter.utils.case_service import (
    CaseFileFormatter,
)
from cg.services.deliver_files.file_formatter.utils.sample_concatenation_service import (
    SampleFileConcatenationFormatter,
)
from cg.services.deliver_files.file_formatter.utils.sample_service import (
    SampleFileFormatter,
)
from cg.store.store import Store


@pytest.fixture
def raw_data_delivery_service(
    delivery_housekeeper_api: HousekeeperAPI,
    delivery_store_microsalt: Store,
) -> RawDataDeliveryFileFetcher:
    """Fixture to get an instance of FetchFastqDeliveryFilesService."""
    tag_service = SampleAndCaseDeliveryTagsFetcher()
    return RawDataDeliveryFileFetcher(
        hk_api=delivery_housekeeper_api,
        status_db=delivery_store_microsalt,
        tags_fetcher=tag_service,
    )


@pytest.fixture
def raw_data_delivery_service_no_housekeeper_bundle(
    real_housekeeper_api: HousekeeperAPI,
    delivery_store_microsalt: Store,
) -> RawDataDeliveryFileFetcher:
    """Fixture to get an instance of FetchFastqDeliveryFilesService."""
    tag_service = SampleAndCaseDeliveryTagsFetcher()
    return RawDataDeliveryFileFetcher(
        hk_api=real_housekeeper_api,
        status_db=delivery_store_microsalt,
        tags_fetcher=tag_service,
    )


@pytest.fixture
def bam_data_delivery_service(
    delivery_housekeeper_api: HousekeeperAPI,
    delivery_store_microsalt: Store,
) -> RawDataDeliveryFileFetcher:
    """Fixture to get an instance of FetchFastqDeliveryFilesService."""
    tag_service = BamDeliveryTagsFetcher()
    return RawDataDeliveryFileFetcher(
        hk_api=delivery_housekeeper_api,
        status_db=delivery_store_microsalt,
        tags_fetcher=tag_service,
    )


@pytest.fixture
def bam_data_delivery_service_no_housekeeper_bundle(
    real_housekeeper_api: HousekeeperAPI,
    delivery_store_microsalt: Store,
) -> RawDataDeliveryFileFetcher:
    """Fixture to get an instance of FetchFastqDeliveryFilesService."""
    tag_service = BamDeliveryTagsFetcher()
    return RawDataDeliveryFileFetcher(
        hk_api=real_housekeeper_api,
        status_db=delivery_store_microsalt,
        tags_fetcher=tag_service,
    )


@pytest.fixture
def analysis_delivery_service(
    delivery_housekeeper_api: HousekeeperAPI,
    delivery_store_balsamic: Store,
) -> AnalysisDeliveryFileFetcher:
    """Fixture to get an instance of FetchAnalysisDeliveryFilesService."""
    tag_service = SampleAndCaseDeliveryTagsFetcher()
    return AnalysisDeliveryFileFetcher(
        hk_api=delivery_housekeeper_api,
        status_db=delivery_store_balsamic,
        tags_fetcher=tag_service,
    )


@pytest.fixture
def analysis_delivery_service_no_housekeeper_bundle(
    real_housekeeper_api: HousekeeperAPI,
    delivery_store_balsamic: Store,
) -> AnalysisDeliveryFileFetcher:
    """Fixture to get an instance of FetchAnalysisDeliveryFilesService."""
    tag_service = SampleAndCaseDeliveryTagsFetcher()
    return AnalysisDeliveryFileFetcher(
        hk_api=real_housekeeper_api,
        status_db=delivery_store_balsamic,
        tags_fetcher=tag_service,
    )


@pytest.fixture
def generic_delivery_file_formatter() -> DeliveryFileFormatter:
    """Fixture to get an instance of GenericDeliveryFileFormatter."""
    return DeliveryFileFormatter(
        sample_file_formatter=SampleFileFormatter(), case_file_formatter=CaseFileFormatter()
    )



""" Trailblazer API for cg."""

import datetime
import logging
from typing import Any

from google.auth import jwt
from google.auth.crypt import RSASigner

from cg.apps.tb.dto.create_job_request import CreateJobRequest
from cg.apps.tb.dto.summary_response import AnalysisSummary, SummariesResponse
from cg.apps.tb.models import AnalysesResponse, TrailblazerAnalysis
from cg.constants import Workflow
from cg.constants.constants import APIMethods, FileFormat, JobType, WorkflowManager
from cg.constants.priority import SlurmQos
from cg.constants.tb import AnalysisStatus
from cg.exc import (
    AnalysisNotCompletedError,
    TrailblazerAnalysisNotFound,
    TrailblazerAPIHTTPError,
)
from cg.io.controller import APIRequest, ReadStream

LOG = logging.getLogger(__name__)


class TrailblazerAPI:
    """Interface to Trailblazer for `cg`."""

    __STARTED_STATUSES = [
        AnalysisStatus.COMPLETED,
        AnalysisStatus.FAILED,
        AnalysisStatus.PENDING,
        AnalysisStatus.RUNNING,
        AnalysisStatus.ERROR,
        AnalysisStatus.QC,
    ]
    __ONGOING_STATUSES = [
        AnalysisStatus.PENDING,
        AnalysisStatus.RUNNING,
        AnalysisStatus.ERROR,
        AnalysisStatus.QC,
    ]

    def __init__(self, config: dict):
        self.service_account = config["trailblazer"]["service_account"]
        self.service_account_auth_file = config["trailblazer"]["service_account_auth_file"]
        self.host = config["trailblazer"]["host"]

    @property
    def auth_header(self) -> dict:
        signer = RSASigner.from_service_account_file(self.service_account_auth_file)
        payload = {"email": self.service_account}
        jwt_token = jwt.encode(signer=signer, payload=payload).decode("ascii")
        return {"Authorization": f"Bearer {jwt_token}"}

    def query_trailblazer(
        self, command: str, request_body: dict, method: str = APIMethods.POST
    ) -> Any:
        url = f"{self.host}/{command}"
        LOG.debug(f"REQUEST HEADER {self.auth_header}")
        LOG.debug(f"{method}: URL={url}; JSON={request_body}")

        response = APIRequest.api_request_from_content(
            api_method=method, url=url, headers=self.auth_header, json=request_body
        )

        LOG.debug(f"RESPONSE STATUS CODE {response.status_code}")
        if not response.ok:
            raise TrailblazerAPIHTTPError(
                f"Request {command} failed with status code {response.status_code}: {response.text}"
            )
        LOG.debug(f"RESPONSE BODY {response.text}")
        return ReadStream.get_content_from_stream(file_format=FileFormat.JSON, stream=response.text)

    def get_latest_completed_analysis(self, case_id: str) -> TrailblazerAnalysis | None:
        endpoint = f"analyses?case_id={case_id}&status[]={AnalysisStatus.COMPLETED}&limit=1"
        response = self.query_trailblazer(command=endpoint, request_body={}, method=APIMethods.GET)
        validated_response = AnalysesResponse.model_validate(response)
        if validated_response.analyses:
            return validated_response.analyses[0]
        raise TrailblazerAnalysisNotFound(f"No completed analysis found for case {case_id}")

    def get_latest_analysis(self, case_id: str) -> TrailblazerAnalysis | None:
        request_body = {
            "case_id": case_id,
        }
        response = self.query_trailblazer(command="get-latest-analysis", request_body=request_body)
        if response:
            return TrailblazerAnalysis.model_validate(response)

    def get_latest_analysis_status(self, case_id: str) -> str | None:
        latest_analysis = self.get_latest_analysis(case_id=case_id)
        if latest_analysis:
            return latest_analysis.status

    def has_latest_analysis_started(self, case_id: str) -> bool:
        return self.get_latest_analysis_status(case_id=case_id) in self.__STARTED_STATUSES

    def is_latest_analysis_ongoing(self, case_id: str) -> bool:
        return self.get_latest_analysis_status(case_id=case_id) in self.__ONGOING_STATUSES

    def is_latest_analysis_completed(self, case_id: str) -> bool:
        return self.get_latest_analysis_status(case_id=case_id) == AnalysisStatus.COMPLETED

    def is_latest_analysis_qc(self, case_id: str) -> bool:
        return self.get_latest_analysis_status(case_id=case_id) == AnalysisStatus.QC

    def add_pending_analysis(
        self,
        case_id: str,
        analysis_type: str,
        config_path: str,
        out_dir: str,
        slurm_quality_of_service: SlurmQos,
        email: str = None,
        order_id: int | None = None,
        workflow: Workflow = None,
        ticket: str = None,
        workflow_manager: str = WorkflowManager.Slurm,
        tower_workflow_id: str | None = None,
        is_hidden: bool = False,
    ) -> TrailblazerAnalysis:
        request_body = {
            "case_id": case_id,
            "email": email,
            "type": analysis_type,
            "config_path": config_path,
            "order_id": order_id,
            "out_dir": out_dir,
            "priority": slurm_quality_of_service,
            "workflow": workflow.upper(),
            "ticket": ticket,
            "workflow_manager": workflow_manager,
            "tower_workflow_id": tower_workflow_id,
            "is_hidden": is_hidden,
        }
        LOG.debug(f"Submitting job to Trailblazer: {request_body}")
        if response := self.query_trailblazer(
            command="add-pending-analysis", request_body=request_body
        ):
            return TrailblazerAnalysis.model_validate(response)

    def set_analysis_uploaded(self, case_id: str, uploaded_at: datetime) -> None:
        """Set a uploaded at date for a trailblazer analysis."""
        request_body = {"case_id": case_id, "uploaded_at": str(uploaded_at)}

        LOG.debug(f"Setting analysis uploaded at for {request_body}")
        LOG.info(f"{case_id} - uploaded at set to {uploaded_at}")
        self.query_trailblazer(
            command="set-analysis-uploaded", request_body=request_body, method=APIMethods.PUT
        )

    def set_analysis_status(self, case_id: str, status: str) -> datetime:
        """Set an analysis to a given status."""
        request_body = {"case_id": case_id, "status": status}

        LOG.debug(f"Request body: {request_body}")
        LOG.info(f"Setting analysis status to {status} for case {case_id}")
        self.query_trailblazer(
            command="set-analysis-status", request_body=request_body, method=APIMethods.PUT
        )

    def add_comment(self, case_id: str, comment: str):
        """Adding a comment to a trailblazer analysis."""
        request_body = {"case_id": case_id, "comment": comment}

        LOG.debug(f"Request body: {request_body}")
        LOG.info(f"Adding comment: {comment} to case {case_id}")
        self.query_trailblazer(
            command="add-comment", request_body=request_body, method=APIMethods.PUT
        )

    def add_upload_job_to_analysis(self, analysis_id: int, slurm_id: int) -> None:
        create_request = CreateJobRequest(slurm_id=slurm_id, job_type=JobType.UPLOAD)
        request_body: dict = create_request.model_dump()
        self.query_trailblazer(
            command=f"/analysis/{analysis_id}/jobs",
            request_body=request_body,
            method=APIMethods.POST,
        )

    def get_summaries(self, order_ids: list[int]) -> list[AnalysisSummary]:
        orders_param = "orderIds=" + ",".join(map(str, order_ids))
        endpoint = f"summary?{orders_param}"
        response = self.query_trailblazer(command=endpoint, request_body={}, method=APIMethods.GET)
        response_data = SummariesResponse.model_validate(response)
        return response_data.summaries

    def get_analyses_to_deliver(self, order_id: int) -> list[TrailblazerAnalysis]:
        """Return the analyses in the order ready to be delivered."""
        endpoint = (
            f"analyses?orderId={order_id}&status[]={AnalysisStatus.COMPLETED}&delivered=false"
        )
        raw_response = self.query_trailblazer(
            command=endpoint, request_body={}, method=APIMethods.GET
        )
        validated_response = AnalysesResponse.model_validate(raw_response)
        return validated_response.analyses

    def verify_latest_analysis_is_completed(self, case_id: str, force: bool = False) -> None:
        """Raises an AnalysisNotCompletedError if the latest analysis for the case has not completed."""
        if not self.is_latest_analysis_completed(case_id) and not force:
            raise AnalysisNotCompletedError(f"The latest analysis for {case_id} has not completed.")



"""Tests for the transfer of external data."""

import logging
from pathlib import Path

from housekeeper.store.models import File
from pytest_mock import MockFixture

from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.constants import SequencingFileTag
from cg.constants.constants import CaseActions
from cg.meta.transfer.external_data import ExternalDataAPI
from cg.store.models import Case, Sample
from cg.store.store import Store
from tests.mocks.hk_mock import MockHousekeeperAPI
from tests.store.conftest import sample_obj


def test_create_log_dir_dry_run(caplog, external_data_api: ExternalDataAPI, ticket_id: str):
    """Test that the directory for logging would have been created."""
    caplog.set_level(logging.INFO)
    # GIVEN an External API with a ticket and dry run set to True
    external_data_api.ticket = ticket_id
    external_data_api.dry_run = True

    # WHEN the log directory is created
    log_dir = external_data_api._create_log_dir()

    # THEN the path is not created since it is a dry run
    assert "Would have created path" in caplog.text

    # THEN the path should start with 2 dirs and then the ticket id
    assert str(log_dir).startswith(f"/another/path/{ticket_id}")


def test_get_source_path(
    cust_sample_id: str,
    external_data_api: ExternalDataAPI,
    ticket_id: str,
):
    """Test generating the source path."""
    # GIVEN an External API with a ticket number, a customer and a customer sample id
    external_data_api.ticket = ticket_id
    customer_id: str = external_data_api.status_db.get_customer_id_from_ticket(ticket=ticket_id)
    external_data_api.customer_id = customer_id

    # WHEN the function is called and assigned
    source_path = external_data_api._get_customer_source_path()

    # THEN the return should be
    assert source_path == Path("server.name.se:/path/cust000/on/caesar/123456/")


def test_get_destination_path(
    customer_id: str,
    external_data_api: ExternalDataAPI,
    sample_id: str,
):
    """Test generating the destination path."""
    # GIVEN an External API with a customer id and an internal sample id
    external_data_api.customer_id = customer_id
    # WHEN the function creates the destination path
    destination_path = external_data_api._get_destination_path(lims_sample_id=sample_id)

    # THEN the destination path should contain the customer_id, ticket_id and sample_id
    assert destination_path == Path("/path/on/hasta/cust000/ADM1/")


def test_transfer_sample_files_from_source(
    caplog,
    customer_id: str,
    cust_sample_id: str,
    external_data_api: ExternalDataAPI,
    external_data_directory: Path,
    helpers,
    mocker,
    ticket_id: str,
):
    caplog.set_level(logging.INFO)

    # GIVEN a Store with three samples, where only two samples are present in the source folder
    for sample in [f"{cust_sample_id}1", f"{cust_sample_id}2", f"{cust_sample_id}3"]:
        helpers.add_sample(
            store=external_data_api.status_db, name=sample, original_ticket=ticket_id
        )

    mocker.patch.object(Store, "get_customer_id_from_ticket")
    Store.get_customer_id_from_ticket.return_value = customer_id

    mocker.patch.object(ExternalDataAPI, "_get_customer_source_path")
    external_data_api._get_customer_source_path.return_value = external_data_directory

    external_data_api.source_path = str(Path("").joinpath(*external_data_directory.parts[:-2]))
    external_data_api.destination_path_stub = str(
        Path("").joinpath(*external_data_directory.parts[:-1], "%s")
    )

    # WHEN the transfer is initiated
    external_data_api.transfer_sample_files_from_source(ticket=ticket_id, dry_run=True)

    # THEN only the two samples present in the source directory are included in the rsync
    assert str(external_data_directory) in caplog.text


def test_curate_sample_folder(
    case_id: str, customer_id: str, external_data_api: ExternalDataAPI, tmpdir_factory
):
    # GIVEN an External API with a customer id
    external_data_api.force = False
    external_data_api.customer_id = customer_id

    # GIVEN a case with a sample
    case: Case = external_data_api.status_db.get_case_by_internal_id(internal_id=case_id)
    sample: Sample = case.links[0].sample
    assert sample

    # GIVEN that the sample folder has the sample name
    tmp_folder = Path(tmpdir_factory.mktemp(sample.name, numbered=False))

    # WHEN the sample folder is curated
    external_data_api._curate_sample_folder(sample_folder=tmp_folder)

    # THEN the sample folder with the sample name does not exist anymore
    assert not tmp_folder.exists()
    # THEN the sample folder with the sample internal id exists
    assert (tmp_folder.parent / sample.internal_id).exists()


def test_get_sample_ids_from_folder(
    external_data_api: ExternalDataAPI,
    sample: Sample,
    ticket_id: str,
    tmp_path: Path,
):
    """Test getting the sample ids from a folder."""
    # GIVEN a sample folder
    tmp_dir_path = Path(tmp_path, sample.internal_id)
    tmp_dir_path.mkdir()
    # GIVEN an External API with a ticket number
    external_data_api.ticket = ticket_id

    # WHEN getting the sample ids from the folder
    available_samples = external_data_api._get_sample_ids_from_folder(folder=tmp_dir_path.parent)

    # THEN the function should return a list containing the sample object
    assert available_samples == [sample.internal_id]
    tmp_dir_path.rmdir()


def test_get_sample_ids_from_folder_no_samples_available(
    external_data_api: ExternalDataAPI,
    ticket_id: str,
    tmpdir_factory,
):
    # GIVEN an External API with a ticket number
    external_data_api.ticket = ticket_id
    # GIVEN that the empty directory created does not contain any correct folders
    tmp_dir_path: Path = Path(tmpdir_factory.mktemp("not_sample_id", numbered=False))
    # WHEN getting the sample ids from the folder
    available_samples = external_data_api._get_sample_ids_from_folder(folder=tmp_dir_path)
    # THEN the function should return an empty list
    assert available_samples == []


def test_get_fastq_paths_to_add(
    external_data_api: ExternalDataAPI,
    fastq_dir: Path,
    mocker: MockFixture,
):
    """Test that getting fastq files to add finds the files in the origin folder."""

    # GIVEN an External API with a sample destination path that has fastq files
    mocker.patch.object(ExternalDataAPI, "_get_destination_path")
    ExternalDataAPI._get_destination_path.return_value = fastq_dir

    # GIVEN a sample id
    sample_id: str = "EXT01"

    # WHEN getting the fastq paths to add
    fastq_paths: list[Path] = external_data_api._get_fastq_paths_to_add(sample_id=sample_id)

    # THEN the function should return a list of fastq file paths
    assert isinstance(fastq_paths[0], Path)
    assert fastq_paths[0].as_posix().endswith(".fastq.gz")


def test_add_and_include_files_to_bundles(
    external_data_api: ExternalDataAPI, sample: Sample, fastq_file: Path
):
    """Test adding and including files to bundles."""
    # GIVEN an External API

    # GIVEN a sample that does not have any fastq files in Housekeeper
    hk_api: HousekeeperAPI = external_data_api.housekeeper_api
    assert not hk_api.files(tags=set(SequencingFileTag.FASTQ), bundle=sample.internal_id).all()

    # WHEN the fastq file is added and included to Housekeeper
    external_data_api._add_and_include_files_to_bundles(
        fastq_paths=[fastq_file], lims_sample_id=sample.internal_id
    )

    # THEN the files should be in Housekeeper
    file: File = hk_api.files(tags=set(SequencingFileTag.FASTQ), bundle=sample.internal_id).first()
    assert file.path.endswith(fastq_file.name)


def test_start_cases(external_data_api: ExternalDataAPI, case_id: str):
    """Test starting cases."""
    # GIVEN an External API

    # GIVEN a case that has not started
    case_not_started: Case = external_data_api.status_db.get_case_by_internal_id(
        internal_id=case_id
    )
    assert case_not_started.action is not CaseActions.ANALYZE

    # WHEN the cases are started
    external_data_api._start_cases(cases=[case_not_started])

    # THEN the case action should be set to analyze
    case_started: Case = external_data_api.status_db.get_case_by_internal_id(internal_id=case_id)
    assert case_started.action == CaseActions.ANALYZE


def test_add_transfer_to_housekeeper(
    case_id: str,
    external_data_api: ExternalDataAPI,
    fastq_file: Path,
    mocker: MockFixture,
    ticket_id: str,
):
    """Test adding samples from a case to Housekeeper"""
    # GIVEN a Store with a DNA case, which is available for analysis
    case = external_data_api.status_db.get_case_by_internal_id(internal_id=case_id)
    mocker.patch.object(Store, "get_cases_by_ticket_id")
    Store.get_cases_by_ticket_id.return_value = [case]
    samples = [fam_sample.sample for fam_sample in case.links]

    # GIVEN a list of paths and only two samples being available
    mocker.patch.object(ExternalDataAPI, "_get_fastq_paths_to_add")
    ExternalDataAPI._get_fastq_paths_to_add.return_value = [fastq_file]

    mocker.patch.object(MockHousekeeperAPI, "last_version")
    MockHousekeeperAPI.last_version.return_value = None

    mocker.patch.object(MockHousekeeperAPI, "get_files")
    MockHousekeeperAPI.get_files.return_value = []

    mocker.patch.object(Path, "iterdir")
    Path.iterdir.return_value = []

    mocker.patch.object(ExternalDataAPI, "_get_available_sample_ids")
    ExternalDataAPI._get_available_sample_ids.return_value = [
        sample.internal_id for sample in samples[:-1]
    ]

    mocker.patch.object(Store, "set_case_action")
    Store.set_case_action.return_value = None

    # THEN none of the samples should exist in housekeeper
    assert all(
        external_data_api.housekeeper_api.bundle(sample.internal_id) is None for sample in samples
    )

    # WHEN the sample bundles are added to housekeeper
    external_data_api.add_external_data_to_housekeeper(ticket=ticket_id)

    # THEN two sample bundles exist in housekeeper and the file has been added to those bundles
    added_samples = list(external_data_api.housekeeper_api.bundles())
    assert all(
        sample.internal_id in [added_sample.name for added_sample in added_samples]
        for sample in samples[:-1]
    )
    assert all(
        sample.versions[0].files[0].path == str(fastq_file.absolute()) for sample in added_samples
    )
    # THEN the sample that is not available should not exist
    assert samples[-1].internal_id not in [added_sample.name for added_sample in added_samples]



from cg.services.sample_run_metrics_service.dtos import SequencingMetrics
from cg.services.sample_run_metrics_service.utils import create_metrics_dto
from cg.store.models import IlluminaSequencingRun
from cg.store.store import Store


class SampleRunMetricsService:
    def __init__(self, store: Store):
        self.store = store

    def get_metrics(self, flow_cell_name: str) -> list[SequencingMetrics]:
        run: IlluminaSequencingRun = self.store.get_illumina_sequencing_run_by_device_internal_id(
            flow_cell_name
        )
        return create_metrics_dto(run.sample_metrics) if run else []



from enum import Enum

from sqlalchemy.orm import Query

from cg.models.orders.constants import OrderType
from cg.store.models import OrderTypeApplication


def filter_applications_by_order_type(
    order_type_applications: Query, order_type: OrderType, **kwargs
) -> Query:
    """Return application by order type."""
    return order_type_applications.filter(OrderTypeApplication.order_type == order_type)


def apply_order_type_application_filter(
    filter_functions: list[callable],
    order_type_applications: Query,
    order_type: OrderType = None,
) -> Query:
    """Apply filtering functions to the ordertype_applications query and return filtered results."""

    for filter_function in filter_functions:
        order_type_applications: Query = filter_function(
            order_type_applications=order_type_applications,
            order_type=order_type,
        )
    return order_type_applications


class OrderTypeApplicationFilter(Enum):
    """Define OrderTypeApplication filter functions."""

    BY_ORDER_TYPE = filter_applications_by_order_type



from typing import Tuple, Union

from pydantic import BaseModel, EmailStr, Field

from cg.clients.freshdesk.constants import Priority, Source, Status


class TicketCreate(BaseModel):
    """Freshdesk ticket."""

    attachments: list[str | bytes] = Field(default_factory=list)
    email: EmailStr
    email_config_id: int | None = None
    description: str
    name: str
    priority: int = Priority.LOW
    source: int = Source.EMAIL
    status: int = Status.OPEN
    subject: str
    tags: list[str] = []
    type: str | None = None
    custom_fields: dict[str, str | int | float | None] = Field(default_factory=dict)

    def to_multipart_data(self) -> list[Tuple[str, str | int | bytes]]:
        """Custom converter to multipart form data."""
        multipart_data = []

        for field, value in self.model_dump(exclude_none=True).items():
            if isinstance(value, list):
                multipart_data.extend([(f"{field}[]", v) for v in value])
            elif isinstance(value, dict):
                multipart_data.extend([(f"{field}[{k}]", v) for k, v in value.items()])
            else:
                multipart_data.append((field, value))

        return multipart_data


class TicketResponse(BaseModel):
    """Response from Freshdesk"""

    id: int
    description: str
    subject: str
    to_emails: list[str] | None = None
    status: int
    priority: int


class ReplyCreate(BaseModel):
    """Reply to a ticket."""

    ticket_number: str
    body: str

    def to_multipart_data(self) -> list[Tuple[str, Union[str, int, bytes]]]:
        """Custom converter to multipart form data."""
        multipart_data = [
            ("body", self.body),
        ]
        return multipart_data



"""File tags for files in Housekeeper."""

from enum import StrEnum

from cg.constants.constants import Workflow


class AlignmentFileTag(StrEnum):
    """Tags for alignment files."""

    BAM: str = "bam"
    BAM_BAI: str = "bai"
    BAM_INDEX: str = "bam-index"
    CRAM: str = "cram"
    CRAM_CRAI: str = "crai"
    CRAM_INDEX: str = "cram-index"

    @classmethod
    def file_tags(cls) -> list[str]:
        return list(map(lambda tag: tag.value, cls))


class ArchiveTag(StrEnum):
    """Tags for archived status."""

    ARCHIVED: str = "archived"


class ScoutTag(StrEnum):
    """Tags for Scout."""

    ARCHIVED: str = "archived"
    SOLVED: str = "solved"


class SequencingFileTag(StrEnum):
    """Tags for sequencing files."""

    CGSTATS_LOG: str = "log"
    DEMUX_LOG: str = "log"
    FASTQ: str = "fastq"
    RUN_PARAMETERS: str = "run-parameters"
    SAMPLE_SHEET: str = "samplesheet"
    SPRING: str = "spring"
    SPRING_METADATA: str = "spring-metadata"


HK_MULTIQC_HTML_TAG = "multiqc-html"

HK_FASTQ_TAGS = [SequencingFileTag.FASTQ]

HK_DELIVERY_REPORT_TAG = "delivery-report"


class HermesFileTag(StrEnum):
    """Tags for hermes."""

    CLINICAL_DELIVERY: str = "clinical-delivery"
    LONG_TERM_STORAGE: str = "long-term-storage"


class AnalysisTag(StrEnum):
    """Tags for analysis files."""

    ARRIBA: str = "arriba"
    ARRIBA_VISUALIZATION: str = "arriba-visualisation"
    BED: str = "bed"
    BIGWIG: str = "bigwig"
    CLINICAL: str = "clinical"
    COVERAGE: str = "coverage"
    FRASER: str = "fraser"
    FUSION: str = "fusion"
    FUSIONCATCHER: str = "fusioncatcher"
    FUSIONCATCHER_SUMMARY: str = "fusioncatcher-summary"
    FUSIONINSPECTOR: str = "fusioninspector"
    FUSIONINSPECTOR_HTML: str = "fusioninspector-html"
    FUSIONREPORT: str = "fusionreport"
    GENE_COUNTS: str = "gene-counts"
    JUNCTION: str = "junction"
    MULTIQC_HTML: str = "multiqc-html"
    OUTRIDER: str = "outrider"
    RESEARCH: str = "research"
    RNA: str = "rna"
    STARFUSION: str = "star-fusion"
    VCF_FUSION: str = "vcf-fusion"


class HkMipAnalysisTag:
    CONFIG: list[str] = ["mip-config"]
    QC_METRICS: list[str] = ["qc-metrics", "deliverable"]
    SAMPLE_INFO: list[str] = ["sample-info"]


class BalsamicAnalysisTag:
    CONFIG: list[str] = ["balsamic-config"]
    QC_METRICS: list[str] = ["qc-metrics", "deliverable"]


class HkAnalysisMetricsTag:
    QC_METRICS: set[str] = {"qc-metrics", "deliverable"}


class GensAnalysisTag:
    COVERAGE: list[str] = ["gens", "coverage", "bed"]
    FRACSNP: list[str] = ["gens", "fracsnp", "bed"]


class GenotypeAnalysisTag:
    GENOTYPE: str = "genotype"


class BalsamicProtectedTags:
    """Balsamic workflow protected tags by type."""

    QC: list[list[str]] = [
        ["balsamic-config"],
        ["balsamic-dag"],
        ["balsamic-report"],
        ["delivery-report"],
        ["multiqc-html"],
        ["multiqc-json"],
        ["qc-metrics"],
    ]
    VARIANT_CALLERS: list[list[str]] = [
        ["ascatngs"],
        ["visualization"],
        ["cnv-report"],
        ["cnvkit"],
        ["delly"],
        ["germline"],
        ["svdb"],
        ["tnscope"],
        ["tnscope-umi"],
        ["vardict"],
        ["vcf2cytosure"],
    ]


WORKFLOW_PROTECTED_TAGS = {
    Workflow.BALSAMIC: BalsamicProtectedTags.QC + BalsamicProtectedTags.VARIANT_CALLERS,
    Workflow.BALSAMIC_QC: BalsamicProtectedTags.QC,
    Workflow.BALSAMIC_PON: [],
    Workflow.BALSAMIC_UMI: BalsamicProtectedTags.QC + BalsamicProtectedTags.VARIANT_CALLERS,
    Workflow.FLUFFY: ["NIPT_csv", "MultiQC"],
    Workflow.MICROSALT: [
        ["microsalt-log"],
        ["config"],
        ["qc-report", "visualization"],
        ["typing-report", "visualization"],
        ["typing-report", "qc-metrics"],
        ["microsalt-config"],
        ["assembly"],
    ],
    Workflow.MIP_DNA: [
        ["vcf-snv-clinical"],
        ["vcf-clinical"],  # legacy
        ["vcf-snv-research"],
        ["vcf-research"],  # legacy
        ["vcf-str"],
        ["snv-gbcf", "snv-bcf"],
        ["bcf-raw"],  # legacy
        ["bcf-raw-index"],  # legacy
        ["mip-config"],
        ["config"],  # legacy
        ["mip-log"],
        ["log"],  # legacy
        ["mip-analyse", "reference-info"],
        ["sample-info"],
        ["sampleinfo"],  # legacy
        ["qc-metrics"],
        ["qc"],  # legacy
        ["smn-calling"],
        ["sv-bcf"],
        ["bcf-raw-sv"],  # legacy
        ["vcf-sv-clinical"],
        ["vcf-clinical-sv"],  # legacy
        ["vcf-clinical-sv-bin"],  # legacy
        ["vcf-clinical-sv-bin-index"],  # legacy
        ["vcf-sv-research"],
        ["vcf-research-sv"],  # legacy
        ["vcf-research-sv-bin"],  # legacy
        ["vcf-research-sv-bin-index"],  # legacy
        ["exe-ver"],
        ["vcf2cytosure"],
        ["delivery-report"],
        ["madeline"],
        ["multiqc-html"],
        ["storage"],
    ],
    Workflow.MIP_RNA: [
        ["vcf-snv-clinical"],
        ["vcf-snv-research"],
        ["mip-config"],
        ["mip-log"],
        ["mip-analyse, reference-info"],
        ["sample-info"],
        ["qc-metrics"],
        ["exe-ver"],
        ["fusion", "vcf"],
        ["salmon-quant"],
    ],
    Workflow.MUTANT: [
        ["fohm-delivery", "instrument-properties"],
        ["fohm-delivery", "pangolin-typing-fohm", "csv"],
        ["vcf", "vcf-report", "fohm-delivery"],
        ["mutant-log"],
        ["metrics"],
        ["config"],
        ["mutant-config"],
        ["software-versions"],
        ["fohm-delivery", "komplettering", "visualization"],
        ["fohm-delivery", "pangolin-typing", "csv", "visualization"],
        ["ks-delivery", "ks-results", "typing-report", "visualization", "csv"],
        ["ks-delivery", "ks-aux-results", "typing-report", "visualization", "csv"],
        ["multiqc-json"],
        ["gisaid-log"],
        ["gisaid-csv"],
    ],
    Workflow.RAREDISEASE: [
        [HermesFileTag.LONG_TERM_STORAGE],
    ],
    Workflow.RAW_DATA: [],
    Workflow.RNAFUSION: [
        [HermesFileTag.LONG_TERM_STORAGE],
        [AnalysisTag.FUSION, AnalysisTag.ARRIBA],  # legacy
        [AnalysisTag.FUSION, AnalysisTag.STARFUSION],  # legacy
        [AnalysisTag.FUSION, AnalysisTag.FUSIONCATCHER],  # legacy
        [AnalysisTag.FUSIONINSPECTOR],  # legacy
        [AnalysisTag.FUSIONREPORT, AnalysisTag.RESEARCH],  # legacy
        [AnalysisTag.FUSIONINSPECTOR_HTML, AnalysisTag.RESEARCH],  # legacy
        [AnalysisTag.ARRIBA_VISUALIZATION, AnalysisTag.RESEARCH],  # legacy
        [AnalysisTag.MULTIQC_HTML, AnalysisTag.RNA],  # legacy
        [HK_DELIVERY_REPORT_TAG],  # legacy
        [AnalysisTag.VCF_FUSION],  # legacy
        [AnalysisTag.GENE_COUNTS],  # legacy
    ],
    Workflow.TAXPROFILER: [
        [HermesFileTag.LONG_TERM_STORAGE],
    ],
    Workflow.TOMTE: [
        [HermesFileTag.LONG_TERM_STORAGE],
    ],
}


class JanusTags:
    """Tags to communicate with the JanusAPI."""

    tags_to_retrieve: list[str] = ["qc-metrics", "janus"]
    multi_qc_file_tags: list[str] = [
        "picard-alignment",
        "picard-duplicates",
        "picard-hs",
        "picard-insert-size",
        "picard-wgs",
        "samtools-stats",
        "somalier",
        "picard-rnaseq",
        "fastp",
        "star",
        "general-stats",
    ]


class FohmTag(StrEnum):
    COMPLEMENTARY = "komplettering"
    PANGOLIN_TYPING = "pangolin-typing-fohm"



import pytest
from unittest.mock import MagicMock

from cg.constants import Workflow, DataDelivery
from cg.services.deliver_files.deliver_files_service.deliver_files_service import (
    DeliverFilesService,
)
from cg.services.deliver_files.deliver_files_service.deliver_files_service_factory import (
    DeliveryServiceFactory,
)
from cg.services.deliver_files.tag_fetcher.sample_and_case_service import (
    SampleAndCaseDeliveryTagsFetcher,
)
from cg.services.deliver_files.file_fetcher.analysis_service import (
    AnalysisDeliveryFileFetcher,
)
from cg.services.deliver_files.file_fetcher.analysis_raw_data_service import (
    RawDataAndAnalysisDeliveryFileFetcher,
)
from cg.services.deliver_files.file_fetcher.raw_data_service import (
    RawDataDeliveryFileFetcher,
)
from cg.services.deliver_files.file_formatter.utils.sample_concatenation_service import (
    SampleFileConcatenationFormatter,
)
from cg.services.deliver_files.file_formatter.utils.sample_service import (
    SampleFileFormatter,
)
from cg.services.deliver_files.file_mover.service import (
    DeliveryFilesMover,
)


@pytest.mark.parametrize(
    "workflow,delivery_type,expected_tag_fetcher,expected_file_fetcher,expected_file_mover,expected_sample_file_formatter",
    [
        (
            Workflow.MICROSALT,
            DataDelivery.FASTQ,
            SampleAndCaseDeliveryTagsFetcher,
            RawDataDeliveryFileFetcher,
            DeliveryFilesMover,
            SampleFileConcatenationFormatter,
        ),
        (
            Workflow.MUTANT,
            DataDelivery.ANALYSIS_FILES,
            SampleAndCaseDeliveryTagsFetcher,
            AnalysisDeliveryFileFetcher,
            DeliveryFilesMover,
            SampleFileFormatter,
        ),
        (
            Workflow.BALSAMIC,
            DataDelivery.FASTQ_ANALYSIS,
            SampleAndCaseDeliveryTagsFetcher,
            RawDataAndAnalysisDeliveryFileFetcher,
            DeliveryFilesMover,
            SampleFileFormatter,
        ),
    ],
)
def test_build_delivery_service(
    workflow,
    delivery_type,
    expected_tag_fetcher,
    expected_file_fetcher,
    expected_file_mover,
    expected_sample_file_formatter,
):
    # GIVEN a delivery service builder with mocked store and hk_api
    builder = DeliveryServiceFactory(
        store=MagicMock(),
        hk_api=MagicMock(),
        rsync_service=MagicMock(),
        tb_service=MagicMock(),
        analysis_service=MagicMock(),
    )

    # WHEN building a delivery service
    delivery_service: DeliverFilesService = builder.build_delivery_service(
        workflow=workflow, delivery_type=delivery_type
    )

    # THEN the correct file formatter and file fetcher services are used
    assert isinstance(delivery_service.file_manager.tags_fetcher, expected_tag_fetcher)
    assert isinstance(delivery_service.file_manager, expected_file_fetcher)
    assert isinstance(delivery_service.file_mover, expected_file_mover)
    assert isinstance(
        delivery_service.file_formatter.sample_file_formatter, expected_sample_file_formatter
    )



from abc import abstractmethod, ABC

from cg.services.deliver_files.file_fetcher.models import DeliveryFiles
from cg.services.deliver_files.file_formatter.models import FormattedFiles


class DeliveryFileFormattingService(ABC):
    """
    Abstract class that encapsulates the logic required for formatting files to deliver.
    """

    @abstractmethod
    def format_files(self, delivery_files: DeliveryFiles) -> FormattedFiles:
        """Format the files to deliver."""
        pass



from cg.services.delivery_message.messages.delivery_message import DeliveryMessage
from cg.services.delivery_message.messages.utils import get_caesar_delivery_path
from cg.store.models import Case


def get_cases_message(cases: list[Case]) -> str:
    delivery_path: str = get_caesar_delivery_path(cases[0])
    return (
        f"Hello,\n\n"
        f"The analysis files are currently being uploaded to your inbox on Caesar:\n\n"
        f"{delivery_path}"
    )


class AnalysisMessage(DeliveryMessage):
    def create_message(self, cases: list[Case]) -> str:
        return get_cases_message(cases)



"""Module for a generic order submitter."""

from cg.models.orders.order import OrderIn
from cg.services.orders.store_order_services.store_case_order import StoreCaseOrderService
from cg.services.orders.submitters.order_submitter import OrderSubmitter
from cg.services.orders.validate_order_services.validate_case_order import (
    ValidateCaseOrderService,
)


class CaseOrderSubmitter(OrderSubmitter):
    """
    Class for submitting generic orders.
    This class is used to submit orders for the following workflows:
    - Balsamic
    - Balsamic QC
    - Balsamic UMI
    - MIP DNA
    - MIP RNA
    - Tomte
    """

    def __init__(
        self,
        order_validation_service: ValidateCaseOrderService,
        order_store_service: StoreCaseOrderService,
    ):
        self.order_validation_service = order_validation_service
        self.order_store_service = order_store_service

    def submit_order(self, order_in: OrderIn) -> dict:
        """Submit a generic order."""
        self.order_validation_service.validate_order(order_in)
        return self.order_store_service.store_order(order_in)



from pathlib import Path

import pytest

from cg.constants.housekeeper_tags import AlignmentFileTag
from cg.constants.pacbio import PacBioHousekeeperTags
from cg.services.run_devices.pacbio.housekeeper_service.models import PacBioFileData


@pytest.fixture
def ccs_report_pac_bio_file_data(
    pacbio_barcoded_ccs_report_file: Path, barcoded_smrt_cell_internal_id: str
) -> PacBioFileData:
    """Return a PacBioFileData object for a CCS report file."""
    return PacBioFileData(
        file_path=pacbio_barcoded_ccs_report_file,
        bundle_name=barcoded_smrt_cell_internal_id,
        tags=[PacBioHousekeeperTags.CCS_REPORT, barcoded_smrt_cell_internal_id],
    )


@pytest.fixture
def pac_bio_hifi_reads_file_data(
    pacbio_barcoded_hifi_read_file: Path,
    barcoded_smrt_cell_internal_id: str,
    pacbio_barcoded_sample_internal_id: str,
) -> PacBioFileData:
    """Return a PacBioFileData object for a HiFi reads file."""
    return PacBioFileData(
        file_path=pacbio_barcoded_hifi_read_file,
        bundle_name=pacbio_barcoded_sample_internal_id,
        tags=[
            AlignmentFileTag.BAM,
            barcoded_smrt_cell_internal_id,
            pacbio_barcoded_sample_internal_id,
        ],
    )



import logging

from cg.meta.delivery_report.balsamic import BalsamicDeliveryReportAPI
from cg.meta.workflow.balsamic_umi import BalsamicUmiAnalysisAPI

LOG = logging.getLogger(__name__)


class BalsamicUmiReportAPI(BalsamicDeliveryReportAPI):
    """API to create Balsamic UMI delivery reports."""

    def __init__(self, analysis_api: BalsamicUmiAnalysisAPI):
        super().__init__(analysis_api=analysis_api)
        self.analysis_api: BalsamicUmiAnalysisAPI = analysis_api



"""Tests delivery report data validation helpers."""

import math

from cg.constants import NA_FIELD
from cg.meta.delivery_report.data_validators import (
    get_empty_report_data,
    get_missing_report_data,
    get_million_read_pairs,
)
from cg.meta.delivery_report.raredisease import RarediseaseDeliveryReportAPI
from cg.models.delivery_report.report import ReportModel
from cg.store.models import Case


def test_get_empty_report_data(
    raredisease_delivery_report_api: RarediseaseDeliveryReportAPI,
    raredisease_case_id: str,
    sample_id: str,
):
    """Test extraction of empty report fields."""

    # GIVEN a delivery report API

    # GIVEN a case object
    case: Case = raredisease_delivery_report_api.analysis_api.status_db.get_case_by_internal_id(
        raredisease_case_id
    )

    # GIVEN a sample ID

    # GIVEN a delivery report data model
    report_data: ReportModel = raredisease_delivery_report_api.get_report_data(
        case_id=raredisease_case_id, analysis_date=case.analyses[0].started_at
    )

    # GIVEN empty fields
    report_data.version = None
    report_data.accredited = None
    report_data.customer.id = NA_FIELD
    report_data.customer.scout_access = False
    report_data.case.samples[0].methods.library_prep = ""
    report_data.case.samples[0].metadata.million_read_pairs = None
    report_data.case.samples[0].metadata.duplicates = None

    # WHEN retrieving the missing data
    empty_report_data: dict = get_empty_report_data(report_data)

    # THEN assert that the empty fields are correctly retrieved
    assert "version" in empty_report_data["report"]
    assert "accredited" in empty_report_data["report"]
    assert "id" in empty_report_data["customer"]
    assert "scout_access" not in empty_report_data["customer"]
    assert "library_prep" in empty_report_data["methods"][sample_id]
    assert "million_read_pairs" in empty_report_data["metadata"][sample_id]
    assert "duplicates" in empty_report_data["metadata"][sample_id]


def test_get_missing_report_data(
    raredisease_delivery_report_api: RarediseaseDeliveryReportAPI,
    raredisease_case_id: str,
    sample_id: str,
):
    """Test getting missing report fields."""

    # GIVEN a delivery report API

    # GIVEN a case object
    case: Case = raredisease_delivery_report_api.analysis_api.status_db.get_case_by_internal_id(
        raredisease_case_id
    )

    # GIVEN a sample ID

    # GIVEN a report data model
    report_data: ReportModel = raredisease_delivery_report_api.get_report_data(
        raredisease_case_id, case.analyses[0].started_at
    )

    # GIVEN a dictionary of report empty fields and a list of required MIP DNA report fields
    empty_fields = {
        "report": ["version", "accredited"],
        "customer": ["id"],
        "methods": {sample_id: ["library_prep"]},
        "metadata": {
            sample_id: ["bait_set", "million_read_pairs", "duplicates"],
        },
    }

    required_fields: dict = raredisease_delivery_report_api.get_required_fields(report_data.case)

    # WHEN retrieving the missing data
    missing_fields: dict = get_missing_report_data(
        empty_fields=empty_fields, required_fields=required_fields
    )

    # THEN assert that the required fields are identified
    assert "version" not in missing_fields["report"]
    assert "accredited" in missing_fields["report"]
    assert "customer" not in missing_fields
    assert "methods" not in missing_fields
    assert "million_read_pairs" in missing_fields["metadata"][sample_id]
    assert "duplicates" in missing_fields["metadata"][sample_id]
    assert "bait_set" not in missing_fields["metadata"][sample_id]


def test_get_million_read_pairs():
    """Test millions read pairs calculation."""

    # GIVEN a number of sequencing reads and its representation in millions of read pairs
    sample_reads = 1_200_000_000
    expected_million_read_pairs = sample_reads / 2_000_000

    # WHEN obtaining the number of reds in millions of read pairs
    million_read_pairs: float = get_million_read_pairs(sample_reads)

    # THEN the expected value should match the calculated one
    assert math.isclose(million_read_pairs, expected_million_read_pairs)


def test_get_million_read_pairs_zero_input():
    """Tests millions read pairs calculation when the sample reads number is zero."""

    # GIVEN zero as number of reads
    sample_reads = 0

    # WHEN retrieving the number of reds in millions of read pairs
    million_read_pairs: float = get_million_read_pairs(sample_reads)

    # THEN the obtained value should be zero
    assert math.isclose(million_read_pairs, 0.0)



"""Tests delivery report models validators."""

from datetime import datetime
from pathlib import Path
from typing import Any

import pytest
from _pytest.logging import LogCaptureFixture
from pydantic_core.core_schema import ValidationInfo
from pytest_mock import MockFixture

from cg.apps.lims import LimsAPI
from cg.constants import (
    NA_FIELD,
    YES_FIELD,
    NO_FIELD,
    REPORT_SEX,
    Sex,
    RIN_MAX_THRESHOLD,
    RIN_MIN_THRESHOLD,
)
from cg.constants.constants import AnalysisType, Workflow
from cg.meta.delivery_report.delivery_report_api import DeliveryReportAPI
from cg.meta.delivery_report.rnafusion import RnafusionDeliveryReportAPI
from cg.models.analysis import NextflowAnalysis
from cg.models.delivery.delivery import DeliveryFile
from cg.models.delivery_report.metadata import RnafusionSampleMetadataModel
from cg.models.delivery_report.validators import (
    get_report_string,
    get_boolean_as_string,
    get_number_as_string,
    get_float_as_percentage,
    get_date_as_string,
    get_list_as_string,
    get_delivered_files_as_file_names,
    get_path_as_string,
    get_sex_as_string,
    get_prep_category_as_string,
    get_analysis_type_as_string,
)
from cg.models.orders.constants import OrderType
from cg.store.models import Case, Analysis, Sample


def test_get_report_string():
    """Test formatting an empty value."""

    # GIVEN a not valid empty field
    none_field = None

    # WHEN performing the validation
    output: str = get_report_string(none_field)

    # THEN check if the input value was formatted correctly
    assert output == NA_FIELD


def test_get_boolean_as_string():
    """Test boolean formatting for the delivery report."""

    # GIVEN a not formatted inputs
    none_field = None
    true_field = True
    false_field = False
    not_a_bool_field = "not a boolean"

    # WHEN performing the validation
    validated_none_field: str = get_boolean_as_string(none_field)
    validated_true_field: str = get_boolean_as_string(true_field)
    validated_false_field: str = get_boolean_as_string(false_field)
    validated_not_bool_field = get_boolean_as_string(not_a_bool_field)

    # THEN check if the input values were formatted correctly
    assert validated_none_field == NA_FIELD
    assert validated_true_field == YES_FIELD
    assert validated_false_field == NO_FIELD
    assert validated_not_bool_field == NA_FIELD


@pytest.mark.parametrize(
    "input_value, expected_output",
    [
        (12.3456789, "12.35"),  # Test for a valid float input
        (0.0, "0.0"),  # Test for float zero input
        (5, "5.0"),  # Test for a valid integer input
        (0, "0.0"),  # Test for integer zero input
        (None, NA_FIELD),  # Test for None input
        ("1.2", "1.2"),  # Test for valid string input
        ("invalid", ValueError),  # Test for an invalid string input
    ],
)
def test_get_number_as_string(input_value: Any, expected_output: str, caplog: LogCaptureFixture):
    """Test the validation and formatting of numbers."""

    # GIVEN a list of number inputs and their expected values

    if expected_output == ValueError:
        # WHEN getting a string representation of a number
        with pytest.raises(ValueError):
            get_number_as_string(input_value)

        # THEN a ValueError should have been raised for an invalid number input
        assert f"Value {input_value} cannot be converted to float" in caplog.text
    else:
        # WHEN getting a string representation of a number
        validated_float_value: str = get_number_as_string(input_value)

        # THEN the expected output should be correctly formatted
        assert validated_float_value == expected_output


def test_get_float_as_percentage():
    """Test the validation of a percentage value."""

    # GIVEN a not formatted percentage
    pct_value = 0.9876

    # WHEN performing the validation
    validated_pct_value: str = get_float_as_percentage(pct_value)

    # THEN check if the input value was formatted correctly
    assert validated_pct_value == "98.76"


def test_get_float_as_percentage_zero_input():
    """Test the validation of a percentage value when input is zero."""

    # GIVEN a zero input
    pct_value = 0.0

    # WHEN performing the validation
    validated_pct_value: str = get_float_as_percentage(pct_value)

    # THEN check if the input value was formatted correctly
    assert validated_pct_value == "0.0"


def test_get_date_as_string(timestamp_now: datetime):
    """Test the validation of a datetime object."""

    # GIVEN a datetime object

    # WHEN performing the validation
    validated_date_value: str = get_date_as_string(timestamp_now)

    # THEN check if the input values were formatted correctly
    assert validated_date_value == str(timestamp_now.date())


def test_get_list_as_string():
    """Test if a list is transformed into a string of comma separated values."""

    # GIVEN a mocked list
    mocked_list = ["I am", "a", "list"]

    # WHEN performing the validation
    validated_list: str = get_list_as_string(mocked_list)

    # THEN check if the input values were formatted correctly
    assert validated_list == "I am, a, list"


def test_get_list_paths_as_strings(filled_file: Path):
    """Test file path name extraction from a list."""

    # GIVEN a list of delivery files
    path_list: list[DeliveryFile] = [
        DeliveryFile(source_path=filled_file, destination_path=filled_file)
    ]

    # WHEN validating the provided list of delivery paths
    path_name_list: list[str] = get_delivered_files_as_file_names(path_list)

    # THEN the returned list should contain the file names
    assert path_name_list[0] == "a_file.txt"


def test_get_path_as_string(filled_file: Path):
    """Test file path name extraction."""

    # GIVEN a mock path

    # WHEN performing the validation
    path_name: str = get_path_as_string(filled_file)

    # THEN check if the input values were formatted correctly
    assert path_name == "a_file.txt"


def test_get_sex_as_string():
    """Test report sex parsing."""

    # GIVEN an invalid sex category
    sex = Sex.FEMALE
    invalid_sex = "not_a_sex"

    # WHEN performing the validation
    validated_sex: str = get_sex_as_string(sex)
    validated_invalid_sex: str = get_sex_as_string(invalid_sex)

    # THEN check if the sex has been correctly formatted
    assert validated_sex == REPORT_SEX.get(sex)
    assert validated_invalid_sex == NA_FIELD


def test_get_prep_category_as_string(caplog: LogCaptureFixture):
    """Test validation on a preparation category."""

    # GIVEN an invalid prep category
    prep_category: OrderType = OrderType.RML

    # WHEN performing the validation

    # THEN check if an exception was raised
    with pytest.raises(ValueError):
        get_prep_category_as_string(prep_category)
    assert "The delivery report generation does not support RML samples" in caplog.text


def test_get_analysis_type_as_string():
    """Test analysis type formatting for the delivery report generation."""

    # GIVEN a WGS analysis type and a model info dictionary
    analysis_type: str = AnalysisType.WHOLE_GENOME_SEQUENCING
    model_info = ValidationInfo
    model_info.data = {"workflow": Workflow.MIP_DNA.value}

    # WHEN performing the validation
    validated_analysis_type: str = get_analysis_type_as_string(
        analysis_type=analysis_type, info=model_info
    )

    # THEN check if the input value was formatted correctly
    assert validated_analysis_type == analysis_type.value


def test_get_analysis_type_as_string_balsamic():
    """Test analysis type formatting for the delivery report generation."""

    # GIVEN a WGS analysis type and a model info dictionary
    analysis_type = "tumor_normal_wgs"
    model_info = ValidationInfo
    model_info.data = {"workflow": Workflow.BALSAMIC.value}

    # WHEN performing the validation
    validated_analysis_type: str = get_analysis_type_as_string(
        analysis_type=analysis_type, info=model_info
    )

    # THEN check if the input value was formatted correctly
    assert validated_analysis_type == "Tumör/normal (helgenomsekvensering)"


def test_check_supported_workflow_mismatch_between_ordered_and_started(
    raredisease_delivery_report_api: DeliveryReportAPI,
    raredisease_case_id: str,
    caplog: LogCaptureFixture,
):
    """Test validation error if a customer requested workflow does not match the data analysis."""

    # GIVEN a delivery report API

    # GIVEN a case object that has been ordered with Raredisease workflow
    case: Case = raredisease_delivery_report_api.analysis_api.status_db.get_case_by_internal_id(
        raredisease_case_id
    )

    # GIVEN an analysis that has been started with the Rnafusion workflow
    analysis: Analysis = case.analyses[0]
    analysis.workflow = Workflow.RNAFUSION

    # WHEN retrieving case analysis data

    # THEN a validation error should be raised
    with pytest.raises(ValueError):
        raredisease_delivery_report_api.get_case_analysis_data(case=case, analysis=analysis)
    assert (
        f"The analysis requested by the customer ({case.data_analysis}) does not match the one executed "
        f"({analysis.workflow})" in caplog.text
    )


def test_check_supported_workflow_not_delivery_report_supported(
    raredisease_delivery_report_api: DeliveryReportAPI,
    raredisease_case_id: str,
    caplog: LogCaptureFixture,
):
    """Test validation error if a customer requested workflow does not match the data analysis."""

    # GIVEN a delivery report API

    # GIVEN a FLUFFY case object
    case: Case = raredisease_delivery_report_api.analysis_api.status_db.get_case_by_internal_id(
        raredisease_case_id
    )
    case.data_analysis = Workflow.FLUFFY

    # GIVEN an analysis that has been started with the FLUFFY workflow
    analysis: Analysis = case.analyses[0]
    analysis.workflow = Workflow.FLUFFY

    # WHEN retrieving case analysis data

    # THEN a validation error should be raised
    with pytest.raises(ValueError):
        raredisease_delivery_report_api.get_case_analysis_data(case=case, analysis=analysis)
    assert (
        f"The workflow {case.data_analysis} does not support delivery report generation"
        in caplog.text
    )


@pytest.mark.parametrize(
    "input_rin, expected_rin",
    [
        (RIN_MAX_THRESHOLD, str(float(RIN_MAX_THRESHOLD))),  # Test for a valid integer input
        (RIN_MAX_THRESHOLD + 1, NA_FIELD),  # Test for an integer above the allowed threshold
        (RIN_MIN_THRESHOLD - 1, NA_FIELD),  # Test for an integer below the allowed threshold
        (None, NA_FIELD),  # Test for a None input
    ],
)
def test_ensure_rin_thresholds(
    rnafusion_case_id: str,
    sample_id: str,
    input_rin: int | float,
    expected_rin: str,
    rnafusion_delivery_report_api: RnafusionDeliveryReportAPI,
    rnafusion_mock_analysis_finish: None,
    mocker: MockFixture,
):
    """Test Rnafusion RIN value validation."""

    # GIVEN a Rnafusion case and associated sample
    case: Case = rnafusion_delivery_report_api.status_db.get_case_by_internal_id(
        internal_id=rnafusion_case_id
    )
    sample: Sample = rnafusion_delivery_report_api.status_db.get_sample_by_internal_id(
        internal_id=sample_id
    )

    # GIVEN an analysis metadata object
    analysis_metadata: NextflowAnalysis = (
        rnafusion_delivery_report_api.analysis_api.get_latest_metadata(rnafusion_case_id)
    )

    # GIVEN a specific RIN value
    mocker.patch.object(LimsAPI, "get_sample_rin", return_value=input_rin)

    # WHEN getting the sample metadata
    sample_metadata: RnafusionSampleMetadataModel = (
        rnafusion_delivery_report_api.get_sample_metadata(
            case=case, sample=sample, analysis_metadata=analysis_metadata
        )
    )

    # THEN the sample RIN value should match the expected RIN value
    assert sample_metadata.rin == expected_rin



"""Tests for RNA part of the scout upload API"""

import logging
from typing import Generator, Set

import pytest
from _pytest.logging import LogCaptureFixture
from housekeeper.store.models import File

import cg.store as Store
from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.constants import HK_MULTIQC_HTML_TAG, Workflow
from cg.constants.scout import ScoutCustomCaseReportTags
from cg.constants.sequencing import SequencingMethod
from cg.exc import CgDataError
from cg.meta.upload.scout.uploadscoutapi import RNADNACollection, UploadScoutAPI
from cg.store.models import Case, Sample
from tests.mocks.hk_mock import MockHousekeeperAPI, MockFile
from tests.store_helpers import StoreHelpers


def set_is_tumour_on_case(store: Store, case_id: str, is_tumour: bool):
    for link in store.get_case_by_internal_id(internal_id=case_id).links:
        link.sample.is_tumour = is_tumour


def get_subject_id_from_case(store: Store, case_id: str) -> str:
    for link in store.get_case_by_internal_id(internal_id=case_id).links:
        return link.sample.subject_id


def ensure_two_dna_tumour_matches(
    dna_case_id: str,
    another_sample_id: str,
    helpers: StoreHelpers,
    rna_case_id: str,
    rna_store: Store,
) -> None:
    """Ensures that we have one RNA case that has two matching DNA cases via subject id and tumour state."""
    set_is_tumour_on_case(store=rna_store, case_id=rna_case_id, is_tumour=True)
    subject_id: str = get_subject_id_from_case(store=rna_store, case_id=rna_case_id)
    set_is_tumour_on_case(store=rna_store, case_id=dna_case_id, is_tumour=True)
    dna_extra_case = helpers.ensure_case(
        store=rna_store, customer=rna_store.get_case_by_internal_id(dna_case_id).customer
    )
    another_sample_id = helpers.add_sample(
        store=rna_store,
        application_tag=SequencingMethod.WGS,
        application_type=SequencingMethod.WGS,
        is_tumour=True,
        name=another_sample_id,
        subject_id=subject_id,
    )
    helpers.add_relationship(store=rna_store, sample=another_sample_id, case=dna_extra_case)
    rna_store.session.commit()


def ensure_extra_rna_case_match(
    another_rna_sample_id: str,
    helpers: StoreHelpers,
    rna_case_id: str,
    rna_store: Store,
) -> None:
    """Ensures that we have an extra RNA case that matches by subject_id the existing RNA case and DNA cases."""
    rna_extra_case = helpers.ensure_case(
        store=rna_store,
        data_analysis=Workflow.MIP_RNA,
        customer=rna_store.get_case_by_internal_id(rna_case_id).customer,
    )
    subject_id: str = get_subject_id_from_case(store=rna_store, case_id=rna_case_id)
    another_rna_sample_id = helpers.add_sample(
        store=rna_store,
        application_type=SequencingMethod.WTS,
        is_tumour=False,
        internal_id=another_rna_sample_id,
        subject_id=subject_id,
    )
    helpers.add_relationship(store=rna_store, sample=another_rna_sample_id, case=rna_extra_case)


def test_upload_rna_alignment_file_to_scout(
    caplog: LogCaptureFixture,
    dna_sample_daughter_id: str,
    dna_sample_father_id: str,
    dna_sample_mother_id: str,
    dna_sample_son_id: str,
    mip_rna_analysis_hk_api: HousekeeperAPI,
    rna_case_id: str,
    rna_store: Store,
    upload_scout_api: UploadScoutAPI,
):
    """Test that a RNA case's alignment file can be loaded via a CG CLI command into an already existing DNA case."""
    caplog.set_level(logging.INFO)

    # GIVEN RNA and DNA cases connected via subject ID
    upload_scout_api.status_db = rna_store

    # GIVEN an RNA case with an alignment file stored in HK

    # WHEN running the method to upload RNA files to Scout
    upload_scout_api.upload_rna_alignment_file(case_id=rna_case_id, dry_run=True)

    # THEN the RNA alignment file should have been uploaded to the linked DNA cases in Scout
    assert "Upload RNA alignment CRAM file finished!" in caplog.text
    for dna_case in [
        dna_sample_mother_id,
        dna_sample_father_id,
        dna_sample_daughter_id,
        dna_sample_son_id,
    ]:
        assert dna_case in caplog.text


def test_upload_rna_junctions_to_scout(
    caplog: Generator[LogCaptureFixture, None, None],
    mip_rna_analysis_hk_api: HousekeeperAPI,
    rna_case_id: str,
    rna_store: Store,
    upload_scout_api: UploadScoutAPI,
):
    """Test that A RNA case's gene fusion report and junction splice files for all samples can be loaded via a cg CLI
    command into an already existing DNA case"""

    # GIVEN an existing RNA case with related sample
    # GIVEN an existing DNA case with related sample
    # GIVEN a sample in the RNA case is connected to a sample in the DNA case via subject_id (i.e. same subject_id)
    upload_scout_api.status_db = rna_store

    # GIVEN the connected RNA sample has a junction bed in Housekeeper
    # GIVEN the connected RNA sample has a bigWig in Housekeeper

    # WHEN running the method to upload RNA files to Scout
    caplog.set_level(logging.INFO)
    upload_scout_api.upload_rna_junctions_to_scout(case_id=rna_case_id, dry_run=True)

    # THEN the 2 files should have been uploaded to the connected sample on the dna case in scout
    assert "Upload splice junctions bed file finished!" in caplog.text
    assert "Upload RNA coverage bigwig file finished!" in caplog.text


def test_upload_splice_junctions_bed_to_scout(
    caplog: Generator[LogCaptureFixture, None, None],
    dna_sample_son_id: str,
    mip_rna_analysis_hk_api: HousekeeperAPI,
    rna_case_id: str,
    rna_sample_son_id: str,
    rna_store: Store,
    upload_scout_api: UploadScoutAPI,
):
    """Test that A RNA case's junction splice files for all samples can be loaded via a cg CLI
    command into an already existing DNA case"""

    # GIVEN an existing RNA case with related sample
    # GIVEN an existing DNA case with related sample
    # GIVEN a sample in the RNA case is connected to a sample in the DNA case via subject_id (i.e. same subject_id)
    upload_scout_api.status_db = rna_store

    # GIVEN the connected RNA sample has a junction bed in Housekeeper

    # WHEN running the method to upload RNA files to Scout
    caplog.set_level(logging.INFO)
    upload_scout_api.upload_splice_junctions_bed_to_scout(case_id=rna_case_id, dry_run=True)

    # THEN the splice junctions file should have been uploaded to the connected sample on the dna case in scout
    assert "Upload splice junctions bed file finished!" in caplog.text

    # THEN the customers dna samples name should have been mentioned in the logging (and used in the upload)
    dna_customer_sample_name: str = rna_store.get_sample_by_internal_id(
        internal_id=dna_sample_son_id
    ).name
    assert dna_customer_sample_name in caplog.text


def test_upload_rna_coverage_bigwig_to_scout(
    caplog: Generator[LogCaptureFixture, None, None],
    dna_sample_son_id: str,
    mip_rna_analysis_hk_api: HousekeeperAPI,
    rna_case_id: str,
    rna_sample_son_id: str,
    rna_store: Store,
    upload_scout_api: UploadScoutAPI,
):
    """Test that A RNA case's bigWig file for all samples can be loaded via a cg CLI
    command into an already existing DNA case"""

    # GIVEN an existing RNA case with related sample
    # GIVEN an existing DNA case with related sample
    # GIVEN a sample in the RNA case is connected to a sample in the DNA case via subject_id (i.e. same subject_id)
    upload_scout_api.status_db = rna_store

    # GIVEN the connected RNA sample has a bigWig in Housekeeper

    # WHEN running the method to upload RNA files to Scout
    caplog.set_level(logging.INFO)
    upload_scout_api.upload_rna_coverage_bigwig_to_scout(case_id=rna_case_id, dry_run=True)

    # THEN the bigWig file should have been uploaded to the connected sample on the dna case in scout
    assert "Upload RNA coverage bigwig file finished!" in caplog.text

    # THEN the customers dna samples name should have been mentioned in the logging (and used in the upload)
    dna_customer_sample_name: str = rna_store.get_sample_by_internal_id(
        internal_id=dna_sample_son_id
    ).name
    assert dna_customer_sample_name in caplog.text


def test_upload_clinical_rna_fusion_report_to_scout(
    caplog: Generator[LogCaptureFixture, None, None],
    dna_case_id: str,
    dna_sample_son_id: str,
    mip_rna_analysis_hk_api: HousekeeperAPI,
    rna_case_id: str,
    rna_sample_son_id: str,
    rna_store: Store,
    upload_scout_api: UploadScoutAPI,
):
    """Test that A RNA case's clinical fusion report and junction splice files for all samples can be loaded via a cg CLI
    command into an already existing DNA case"""

    # GIVEN an existing RNA case with related sample
    # GIVEN an existing DNA case with related sample
    # GIVEN a sample in the RNA case is connected to a sample in the DNA case via subject_id (i.e. same subject_id)
    upload_scout_api.status_db = rna_store

    # GIVEN the connected RNA case has a clinical fusion report in Housekeeper

    # WHEN running the method to upload RNA files to Scout
    caplog.set_level(logging.INFO)
    upload_scout_api.upload_fusion_report_to_scout(case_id=rna_case_id, dry_run=True)

    # THEN the fusion report file should have been uploaded to the connected sample on the dna case in scout
    assert "Upload Clinical fusion report finished!" in caplog.text

    # THEN the dna case id should have been mentioned in the logging (and used in the upload)
    assert dna_case_id in caplog.text

    # THEN the customers dna samples name should NOT have been mentioned in the logging
    dna_customer_sample_name: str = rna_store.get_sample_by_internal_id(
        internal_id=dna_sample_son_id
    ).name
    assert dna_customer_sample_name not in caplog.text


def test_upload_research_rna_fusion_report_to_scout(
    caplog: Generator[LogCaptureFixture, None, None],
    dna_case_id: str,
    dna_sample_son_id: str,
    mip_rna_analysis_hk_api: HousekeeperAPI,
    rna_case_id: str,
    rna_sample_son_id: str,
    rna_store: Store,
    upload_scout_api: UploadScoutAPI,
):
    """Test that A RNA case's gene fusion report and junction splice files for all samples can be loaded via a cg CLI
    command into an already existing DNA case"""

    # GIVEN an existing RNA case with related sample
    # GIVEN an existing DNA case with related sample
    # GIVEN a sample in the RNA case is connected to a sample in the DNA case via subject_id (i.e. same subject_id)
    upload_scout_api.status_db = rna_store

    # GIVEN the connected RNA case has a research fusion report in Housekeeper

    # WHEN running the method to upload RNA files to Scout
    caplog.set_level(logging.INFO)
    upload_scout_api.upload_fusion_report_to_scout(case_id=rna_case_id, dry_run=True, research=True)

    # THEN the fusion report file should have been uploaded to the connected sample on the dna case in scout
    assert "Upload Research fusion report finished!" in caplog.text

    # THEN the dna case id should have been mentioned in the logging (and used in the upload)
    assert dna_case_id in caplog.text

    # THEN the customers dna samples name should NOT have been mentioned in the logging
    dna_customer_sample_name: str = rna_store.get_sample_by_internal_id(
        internal_id=dna_sample_son_id
    ).name
    assert dna_customer_sample_name not in caplog.text


def test_upload_rna_fusion_report_to_scout_no_subject_id(
    caplog: Generator[LogCaptureFixture, None, None],
    dna_case_id: str,
    mip_rna_analysis_hk_api: HousekeeperAPI,
    rna_case_id: str,
    rna_store: Store,
    upload_scout_api: UploadScoutAPI,
):
    """Test that A RNA case's gene fusion report"""

    # GIVEN a sample in the RNA case is NOT connected to a sample in the DNA case via subject_id (i.e. same subject_id)
    for link in rna_store.get_case_by_internal_id(rna_case_id).links:
        link.sample.subject_id = ""
    for link in rna_store.get_case_by_internal_id(dna_case_id).links:
        link.sample.subject_id = ""
    rna_store.session.commit()
    upload_scout_api.status_db = rna_store

    # GIVEN the connected RNA case has a research fusion report in Housekeeper

    # WHEN running the method to upload RNA files to Scout
    caplog.set_level(logging.INFO)
    # THEN an exception should be raised on unconnected data
    with pytest.raises(CgDataError):
        upload_scout_api.upload_fusion_report_to_scout(case_id=rna_case_id, dry_run=True)


def test_upload_rna_coverage_bigwig_to_scout_no_subject_id(
    caplog: Generator[LogCaptureFixture, None, None],
    dna_case_id: str,
    mip_rna_analysis_hk_api: HousekeeperAPI,
    rna_case_id: str,
    rna_store: Store,
    upload_scout_api: UploadScoutAPI,
):
    """Test that A RNA case's gene fusion report and junction splice files for all samples can be loaded via a cg CLI
    command into an already existing DNA case"""

    # GIVEN a sample in the RNA case is NOT connected to a sample in the DNA case via subject_id (i.e. same subject_id)
    for link in rna_store.get_case_by_internal_id(rna_case_id).links:
        link.sample.subject_id = ""
    for link in rna_store.get_case_by_internal_id(dna_case_id).links:
        link.sample.subject_id = ""
    rna_store.session.commit()
    upload_scout_api.status_db = rna_store

    # GIVEN the connected RNA sample has a bigWig in Housekeeper

    # WHEN running the method to upload RNA files to Scout
    caplog.set_level(logging.INFO)
    # THEN an exception should be raised on unconnected data
    with pytest.raises(CgDataError):
        upload_scout_api.upload_rna_coverage_bigwig_to_scout(case_id=rna_case_id, dry_run=True)


def test_upload_splice_junctions_bed_to_scout_no_subject_id(
    caplog: Generator[LogCaptureFixture, None, None],
    dna_case_id: str,
    mip_rna_analysis_hk_api: HousekeeperAPI,
    rna_case_id: str,
    rna_store: Store,
    upload_scout_api: UploadScoutAPI,
):
    """Test that A RNA case's junction splice files for all samples can be loaded via a cg CLI
    command into an already existing DNA case"""

    # GIVEN a sample in the RNA case is NOT connected to a sample in the DNA case via subject_id (i.e. same subject_id)
    for link in rna_store.get_case_by_internal_id(rna_case_id).links:
        link.sample.subject_id = ""
    for link in rna_store.get_case_by_internal_id(dna_case_id).links:
        link.sample.subject_id = ""
    rna_store.session.commit()
    upload_scout_api.status_db = rna_store

    # GIVEN the connected RNA sample has a junction bed in Housekeeper

    # WHEN running the method to upload RNA files to Scout
    caplog.set_level(logging.INFO)
    # THEN an exception should be raised on unconnected data
    with pytest.raises(CgDataError):
        upload_scout_api.upload_splice_junctions_bed_to_scout(case_id=rna_case_id, dry_run=True)


def test_upload_rna_fusion_report_to_scout_tumour_non_matching(
    caplog: Generator[LogCaptureFixture, None, None],
    dna_case_id: str,
    mip_rna_analysis_hk_api: HousekeeperAPI,
    rna_case_id: str,
    rna_store: Store,
    upload_scout_api: UploadScoutAPI,
):
    """Test that an RNA case's gene fusion report is not uploaded if the is_tumour is not matching"""

    # GIVEN a sample in the RNA case is NOT connected to a sample in the DNA case via is_tumour (i.e. different is_tumour)
    set_is_tumour_on_case(store=rna_store, case_id=rna_case_id, is_tumour=True)
    set_is_tumour_on_case(store=rna_store, case_id=dna_case_id, is_tumour=False)
    rna_store.session.commit()
    upload_scout_api.status_db = rna_store

    # GIVEN the connected RNA case has a research fusion report in Housekeeper

    # WHEN running the method to upload RNA files to Scout
    caplog.set_level(logging.INFO)

    # THEN an exception should be raised on unconnected data
    with pytest.raises(CgDataError):
        upload_scout_api.upload_fusion_report_to_scout(case_id=rna_case_id, dry_run=True)


def test_upload_rna_coverage_bigwig_to_scout_tumour_non_matching(
    caplog: Generator[LogCaptureFixture, None, None],
    dna_case_id: str,
    mip_rna_analysis_hk_api: HousekeeperAPI,
    rna_case_id: str,
    rna_store: Store,
    upload_scout_api: UploadScoutAPI,
):
    """Test that A RNA case's gene fusion report and junction splice files for all samples is not uploaded if the is_tumour is not matching"""

    # GIVEN a sample in the RNA case is NOT connected to a sample in the DNA case via is_tumour (i.e. different is_tumour)
    set_is_tumour_on_case(store=rna_store, case_id=rna_case_id, is_tumour=True)
    set_is_tumour_on_case(store=rna_store, case_id=dna_case_id, is_tumour=False)
    rna_store.session.commit()
    upload_scout_api.status_db = rna_store

    # GIVEN the connected RNA sample has a bigWig in Housekeeper

    # WHEN running the method to upload RNA files to Scout
    caplog.set_level(logging.INFO)

    # THEN an exception should be raised on unconnected data
    with pytest.raises(CgDataError):
        upload_scout_api.upload_rna_coverage_bigwig_to_scout(case_id=rna_case_id, dry_run=True)


def test_upload_splice_junctions_bed_to_scout_tumour_non_matching(
    caplog: Generator[LogCaptureFixture, None, None],
    dna_case_id: str,
    mip_rna_analysis_hk_api: HousekeeperAPI,
    rna_case_id: str,
    rna_store: Store,
    upload_scout_api: UploadScoutAPI,
):
    """Test that A RNA case's junction splice files for all samples is not uploaded if the is_tumour is not matching"""

    # GIVEN a sample in the RNA case is NOT connected to a sample in the DNA case via is_tumour (i.e. different is_tumour)
    set_is_tumour_on_case(store=rna_store, case_id=rna_case_id, is_tumour=True)
    set_is_tumour_on_case(store=rna_store, case_id=dna_case_id, is_tumour=False)
    rna_store.session.commit()
    upload_scout_api.status_db = rna_store

    # GIVEN the connected RNA sample has a junction bed in Housekeeper

    # WHEN running the method to upload RNA files to Scout
    caplog.set_level(logging.INFO)

    # THEN an exception should be raised on unconnected data
    with pytest.raises(CgDataError):
        upload_scout_api.upload_splice_junctions_bed_to_scout(case_id=rna_case_id, dry_run=True)


def test_upload_rna_fusion_report_to_scout_tumour_multiple_matches(
    caplog: Generator[LogCaptureFixture, None, None],
    dna_case_id: str,
    another_sample_id: str,
    helpers: StoreHelpers,
    mip_rna_analysis_hk_api: HousekeeperAPI,
    rna_case_id: str,
    rna_store: Store,
    upload_scout_api: UploadScoutAPI,
):
    """Test that an RNA case's gene fusion report is not uploaded if the is_tumour has too many DNA-matches"""

    # GIVEN a sample in the RNA case is connected to a sample in the DNA case via is_tumour (i.e. same is_tumour)
    ensure_two_dna_tumour_matches(dna_case_id, another_sample_id, helpers, rna_case_id, rna_store)
    upload_scout_api.status_db = rna_store

    # GIVEN the connected RNA case has a research fusion report in Housekeeper

    # WHEN running the method to upload RNA files to Scout
    caplog.set_level(logging.INFO)

    # THEN an exception should be raised on unconnected data
    with pytest.raises(CgDataError):
        upload_scout_api.upload_fusion_report_to_scout(case_id=rna_case_id, dry_run=True)


def test_upload_rna_coverage_bigwig_to_scout_tumour_multiple_matches(
    caplog: Generator[LogCaptureFixture, None, None],
    dna_case_id: str,
    another_sample_id: str,
    helpers: StoreHelpers,
    mip_rna_analysis_hk_api: HousekeeperAPI,
    rna_case_id: str,
    rna_store: Store,
    upload_scout_api: UploadScoutAPI,
):
    """Test that A RNA case's gene fusion report and junction splice files for all samples is not uploaded if the RNA-sample has too many DNA-matches"""

    # GIVEN a sample in the RNA case is connected to a sample in the DNA case via is_tumour (i.e. same is_tumour)
    ensure_two_dna_tumour_matches(dna_case_id, another_sample_id, helpers, rna_case_id, rna_store)
    upload_scout_api.status_db = rna_store

    # GIVEN the connected RNA sample has a bigWig in Housekeeper

    # WHEN running the method to upload RNA files to Scout
    caplog.set_level(logging.INFO)

    # THEN an exception should be raised on unconnected data
    with pytest.raises(CgDataError):
        upload_scout_api.upload_rna_coverage_bigwig_to_scout(case_id=rna_case_id, dry_run=True)


def test_upload_splice_junctions_bed_to_scout_tumour_multiple_matches(
    caplog: Generator[LogCaptureFixture, None, None],
    dna_case_id: str,
    another_sample_id: str,
    helpers: StoreHelpers,
    mip_rna_analysis_hk_api: HousekeeperAPI,
    rna_case_id: str,
    rna_store: Store,
    upload_scout_api: UploadScoutAPI,
):
    """Test that A RNA case's junction splice files for all samples is not uploaded if the RNA-sample has too many DNA-matches"""

    # GIVEN a sample in the RNA case is connected to a sample in the DNA case via is_tumour (i.e. same is_tumour)
    ensure_two_dna_tumour_matches(dna_case_id, another_sample_id, helpers, rna_case_id, rna_store)
    upload_scout_api.status_db = rna_store

    # GIVEN the connected RNA sample has a junction bed in Housekeeper

    # WHEN running the method to upload RNA files to Scout
    caplog.set_level(logging.INFO)

    # THEN an exception should be raised on unconnected data
    with pytest.raises(CgDataError):
        upload_scout_api.upload_splice_junctions_bed_to_scout(case_id=rna_case_id, dry_run=True)


def test_get_application_prep_category(
    another_rna_sample_id: str,
    dna_sample_son_id: str,
    helpers: StoreHelpers,
    rna_case_id: str,
    rna_sample_son_id: str,
    rna_store: Store,
    upload_scout_api: UploadScoutAPI,
):
    """Test that RNA samples are removed when filtering sample list by workflow."""

    # GIVEN an RNA sample that is connected by subject ID to one RNA and one DNA sample in other cases

    ensure_extra_rna_case_match(another_rna_sample_id, helpers, rna_case_id, rna_store)
    upload_scout_api.status_db = rna_store

    dna_sample: Sample = rna_store.get_sample_by_internal_id(dna_sample_son_id)
    another_rna_sample_id: Sample = rna_store.get_sample_by_internal_id(another_rna_sample_id)
    all_son_rna_dna_samples: list[Sample] = [dna_sample, another_rna_sample_id]

    # WHEN running the method to filter a list of Sample objects containing RNA and DNA samples connected by subject_id
    only_son_dna_samples = upload_scout_api._get_application_prep_category(all_son_rna_dna_samples)

    # THEN even though an RNA sample is present in the initial query, the output should not contain any RNA samples
    nr_of_subject_id_samples: int = len(all_son_rna_dna_samples)
    nr_of_subject_id_dna_samples: int = len([only_son_dna_samples])
    assert nr_of_subject_id_samples == 2
    assert nr_of_subject_id_dna_samples == 1


def test_create_rna_dna_collections(
    rna_case_id: str,
    rna_store: Store,
    upload_scout_api: UploadScoutAPI,
):
    """Test that the create_rna_dna_collections returns a list of RNADNACollections."""

    # GIVEN an RNA case with RNA samples that are connected by subject ID to DNA samples in a DNA case
    rna_case: Case = rna_store.get_case_by_internal_id(rna_case_id)

    # WHEN running the method to create a list of RNADNACollections
    # with the relationships between RNA/DNA samples and DNA cases
    rna_dna_collections: list[RNADNACollection] = upload_scout_api.create_rna_dna_collections(
        rna_case
    )

    # THEN the output should be a list of RNADNACollections
    assert all(
        isinstance(rna_dna_collection, RNADNACollection)
        for rna_dna_collection in rna_dna_collections
    )


def test_add_rna_sample(
    rna_case_id: str,
    rna_store: Store,
    upload_scout_api: UploadScoutAPI,
):
    """Test that for a given RNA case the RNA samples are added to the RNADNACollections."""

    # GIVEN an RNA case and the associated RNA samples
    rna_case: Case = rna_store.get_case_by_internal_id(internal_id=rna_case_id)
    rna_sample_list: list[Sample] = (
        rna_store._get_query(table=Sample).filter(Sample.internal_id.contains("rna")).all()
    )

    # WHEN running the method to create a list of RNADNACollections
    # with the relationships between RNA/DNA samples and DNA cases
    rna_dna_collections: list[RNADNACollection] = upload_scout_api.create_rna_dna_collections(
        rna_case
    )

    # THEN the resulting RNADNACollections should contain all RNA samples in the case
    assert rna_sample_list
    for sample in rna_sample_list:
        assert sample.internal_id in [
            rna_dna_collection.rna_sample_internal_id for rna_dna_collection in rna_dna_collections
        ]


def test_link_rna_sample_to_dna_sample(
    dna_sample_son_id: str,
    rna_sample_son_id: str,
    rna_store: Store,
    upload_scout_api: UploadScoutAPI,
):
    """Test for a given RNA sample, the associated DNA sample name matches and is present in rna_dna_case_map."""

    # GIVEN an RNA sample
    rna_sample: Sample = rna_store.get_sample_by_internal_id(rna_sample_son_id)

    # WHEN creating an RNADNACollection for an RNA sample
    rna_dna_collection: RNADNACollection = upload_scout_api.create_rna_dna_collection(rna_sample)

    # THEN the RNADNACollection should contain the RNA sample
    assert rna_sample_son_id == rna_dna_collection.rna_sample_internal_id

    # THEN the RNADNACollection should have its dna_sample_id set to the related dna_sample_id
    assert dna_sample_son_id == rna_dna_collection.dna_sample_name


def test_add_dna_cases_to_dna_sample(
    dna_case_id: str,
    dna_sample_son_id: str,
    rna_sample_son_id: str,
    rna_store: Store,
    upload_scout_api: UploadScoutAPI,
):
    """Test for a given RNA sample, the DNA case name matches to the case name of the DNA sample in the RNADNACollection."""

    # GIVEN an RNA sample, a DNA sample, and a DNA case
    rna_sample: Sample = rna_store.get_sample_by_internal_id(internal_id=rna_sample_son_id)
    dna_case: Case = rna_store.get_case_by_internal_id(internal_id=dna_case_id)

    # WHEN adding creating the RNADNACollection
    rna_dna_collection: RNADNACollection = upload_scout_api.create_rna_dna_collection(rna_sample)

    # THEN the DNA cases should contain the DNA_case name associated with the DNA sample
    assert dna_case.internal_id in rna_dna_collection.dna_case_ids


def test_map_dna_cases_to_dna_sample_incorrect_workflow(
    rna_store: Store,
    upload_scout_api: UploadScoutAPI,
    dna_sample_son_id: str,
    dna_case_id: str,
    rna_sample_son_id: str,
):
    """Test that the DNA case name is not mapped to the DNA sample name in the rna-dna-sample-case map."""

    # GIVEN an RNA sample and a DNA sample

    dna_sample: Sample = rna_store.get_sample_by_internal_id(dna_sample_son_id)
    dna_case: Case = rna_store.get_case_by_internal_id(dna_case_id)
    rna_sample: Sample = rna_store.get_sample_by_internal_id(rna_sample_son_id)

    # GIVEN that the DNA case has a different workflow than the expected workflow
    dna_case.data_analysis = Workflow.RAW_DATA

    # WHEN mapping the DNA case name to the DNA sample name in the related DNA cases
    related_dna_cases: list[str] = upload_scout_api._dna_cases_related_to_dna_sample(
        dna_sample=dna_sample,
        collaborators=rna_sample.customer.collaborators,
    )

    # THEN the related DNA cases should not contain the DNA case name associated with the DNA sample name
    assert dna_case.internal_id not in related_dna_cases


def test_map_dna_cases_to_dna_sample_incorrect_customer(
    rna_store: Store,
    upload_scout_api: UploadScoutAPI,
    dna_sample_son_id: str,
    dna_case_id: str,
    rna_sample_son_id: str,
):
    """Test that the DNA case name is not mapped to the DNA sample name in the RNADNACollection."""

    # GIVEN an RNA sample, a DNA sample, and a rna-dna case map

    dna_sample: Sample = rna_store.get_sample_by_internal_id(internal_id=dna_sample_son_id)
    dna_case: Case = rna_store.get_case_by_internal_id(internal_id=dna_case_id)
    rna_sample: Sample = rna_store.get_sample_by_internal_id(internal_id=rna_sample_son_id)

    # GIVEN that the DNA case has a different customer than the expected customer
    dna_case.customer_id = 1000

    # WHEN mapping the DNA case name to the DNA sample name in the rna-dna-sample-case map
    dna_cases: list[str] = upload_scout_api._dna_cases_related_to_dna_sample(
        dna_sample=dna_sample,
        collaborators=rna_sample.customer.collaborators,
    )

    # THEN the rna-dna-sample-case map should not contain the DNA case name associated with the DNA sample name
    assert dna_case.internal_id not in dna_cases


def test_get_multiqc_html_report(
    dna_case_id: str,
    rna_store: Store,
    upload_mip_analysis_scout_api: UploadScoutAPI,
    mip_dna_analysis_hk_api: MockHousekeeperAPI,
):
    """Test that the multiqc html report is returned."""

    # GIVEN a DNA case with a multiqc-html report
    case: Case = rna_store.get_case_by_internal_id(internal_id=dna_case_id)
    multiqc_file: MockFile = mip_dna_analysis_hk_api.get_file_from_latest_version(
        bundle_name=dna_case_id, tags={HK_MULTIQC_HTML_TAG, case}
    )

    # WHEN getting the multiqc html report
    report_type, multiqc_report = upload_mip_analysis_scout_api.get_multiqc_html_report(
        case_id=dna_case_id, workflow=case.data_analysis
    )

    # THEN the multiqc html report should be returned and the correct report type
    assert multiqc_report.full_path == multiqc_file.full_path
    assert report_type == ScoutCustomCaseReportTags.MULTIQC


def test_upload_report_to_scout(
    caplog,
    dna_case_id: str,
    upload_mip_analysis_scout_api: UploadScoutAPI,
    mip_dna_analysis_hk_api: MockHousekeeperAPI,
):
    """Test that the uploaded of a report to Scout is possible."""

    caplog.set_level(logging.INFO)

    # GIVEN a DNA case with a multiqc-html report
    multiqc_file: File = mip_dna_analysis_hk_api.files(
        bundle=dna_case_id, tags=[ScoutCustomCaseReportTags.MULTIQC]
    )[0]
    assert multiqc_file

    # WHEN uploading a report to Scout
    upload_mip_analysis_scout_api.upload_report_to_scout(
        dry_run=False,
        case_id=dna_case_id,
        report_type=ScoutCustomCaseReportTags.MULTIQC,
        report_file=multiqc_file,
    )

    # THEN the report should be uploaded to Scout
    assert (
        f"Uploading {ScoutCustomCaseReportTags.MULTIQC} report to case {dna_case_id}" in caplog.text
    )


def test_upload_rna_multiqc_report_to_successful_dna_case_in_scout(
    caplog,
    rna_case_id: str,
    rna_store: Store,
    upload_mip_analysis_scout_api: UploadScoutAPI,
    mip_rna_analysis_hk_api: MockHousekeeperAPI,
):
    """Test that the report is uploaded to Scout."""

    caplog.set_level(logging.INFO)

    # GIVEN an RNA case, and an store with an rna connected to it
    upload_mip_analysis_scout_api.status_db: Store = rna_store

    # GIVEN an RNA case with a multiqc-htlml report
    multiqc_file: File = mip_rna_analysis_hk_api.files(
        bundle=rna_case_id, tags=[ScoutCustomCaseReportTags.MULTIQC]
    )[0]

    # WHEN uploading a report to a completed DNA case in Scout
    upload_mip_analysis_scout_api.upload_rna_report_to_dna_case_in_scout(
        dry_run=False,
        rna_case_id=rna_case_id,
        report_type=ScoutCustomCaseReportTags.MULTIQC,
        report_file=multiqc_file,
    )

    # WHEN finding the related DNA case
    dna_case_ids: Set[str] = upload_mip_analysis_scout_api.get_unique_dna_cases_related_to_rna_case(
        case_id=rna_case_id
    )

    # THEN the api should know that it should find related DNA cases
    assert f"Finding DNA cases related to RNA case {rna_case_id}" in caplog.text

    # THEN the report should be uploaded to Scout
    for case_id in dna_case_ids:
        assert (
            f"Uploading {ScoutCustomCaseReportTags.MULTIQC} report to Scout for case {case_id}"
            in caplog.text
        )


def test_upload_tomte_rna_multiqc_report_to_successful_dna_case_in_scout(
    caplog,
    rna_case_id: str,
    rna_store: Store,
    upload_mip_analysis_scout_api: UploadScoutAPI,
    tomte_analysis_hk_api: MockHousekeeperAPI,
):
    """Test that the report is uploaded to Scout."""

    caplog.set_level(logging.INFO)

    # GIVEN an RNA case, and an store with an rna connected to it
    upload_mip_analysis_scout_api.status_db: Store = rna_store

    # GIVEN an RNA case with a multiqc-htlml report
    multiqc_file: File = tomte_analysis_hk_api.files(
        bundle=rna_case_id, tags=[ScoutCustomCaseReportTags.MULTIQC]
    )[0]

    # WHEN uploading a report to a completed DNA case in Scout
    upload_mip_analysis_scout_api.upload_rna_report_to_dna_case_in_scout(
        dry_run=False,
        rna_case_id=rna_case_id,
        report_type=ScoutCustomCaseReportTags.MULTIQC,
        report_file=multiqc_file,
    )

    # WHEN finding the related DNA case
    dna_case_ids: Set[str] = upload_mip_analysis_scout_api.get_unique_dna_cases_related_to_rna_case(
        case_id=rna_case_id
    )

    # THEN the api should know that it should find related DNA cases
    assert f"Finding DNA cases related to RNA case {rna_case_id}" in caplog.text

    # THEN the report should be uploaded to Scout
    for case_id in dna_case_ids:
        assert (
            f"Uploading {ScoutCustomCaseReportTags.MULTIQC} report to Scout for case {case_id}"
            in caplog.text
        )


def test_upload_rna_delivery_report_to_successful_dna_case_in_scout(
    caplog,
    rna_case_id: str,
    rna_store: Store,
    upload_mip_analysis_scout_api: UploadScoutAPI,
    mip_rna_analysis_hk_api: MockHousekeeperAPI,
):
    """Test that the report is uploaded to Scout."""

    caplog.set_level(logging.INFO)

    # GIVEN an RNA case, and an store with an rna connected to it
    upload_mip_analysis_scout_api.status_db: Store = rna_store

    # GIVEN an RNA case with a multiqc-htlml report
    rna_delivery_report: File = mip_rna_analysis_hk_api.files(
        bundle=rna_case_id, tags=[ScoutCustomCaseReportTags.RNA_DELIVERY]
    )[0]

    # WHEN uploading a report to a completed DNA case in Scout
    upload_mip_analysis_scout_api.upload_rna_report_to_dna_case_in_scout(
        dry_run=False,
        rna_case_id=rna_case_id,
        report_type=ScoutCustomCaseReportTags.RNA_DELIVERY,
        report_file=rna_delivery_report,
    )

    # WHEN finding the related DNA case
    dna_case_ids: Set[str] = upload_mip_analysis_scout_api.get_unique_dna_cases_related_to_rna_case(
        case_id=rna_case_id
    )

    # THEN the api should know that it should find related DNA cases
    assert f"Finding DNA cases related to RNA case {rna_case_id}" in caplog.text

    # THEN the report should be uploaded to Scout
    for case_id in dna_case_ids:
        assert (
            f"Uploading {ScoutCustomCaseReportTags.RNA_DELIVERY} report to Scout for case {case_id}"
            in caplog.text
        )


def test_upload_tomte_rna_delivery_report_to_successful_dna_case_in_scout(
    caplog,
    rna_case_id: str,
    rna_store: Store,
    upload_tomte_analysis_scout_api: UploadScoutAPI,
    tomte_analysis_hk_api: MockHousekeeperAPI,
):
    """Test that the report is uploaded to Scout."""

    caplog.set_level(logging.INFO)

    # GIVEN an RNA case, and an store with an rna connected to it
    upload_tomte_analysis_scout_api.status_db: Store = rna_store

    # GIVEN an RNA case with a rna delivery report
    rna_delivery_report: File = tomte_analysis_hk_api.files(
        bundle=rna_case_id, tags=[ScoutCustomCaseReportTags.RNA_DELIVERY]
    )[0]

    # WHEN uploading a report to a completed DNA case in Scout
    upload_tomte_analysis_scout_api.upload_rna_report_to_dna_case_in_scout(
        dry_run=False,
        rna_case_id=rna_case_id,
        report_type=ScoutCustomCaseReportTags.RNA_DELIVERY,
        report_file=rna_delivery_report,
    )

    # WHEN finding the related DNA case
    dna_case_ids: Set[str] = (
        upload_tomte_analysis_scout_api.get_unique_dna_cases_related_to_rna_case(
            case_id=rna_case_id
        )
    )

    dna_case_ids: Set[str] = (
        upload_tomte_analysis_scout_api.get_unique_dna_cases_related_to_rna_case(
            case_id=rna_case_id
        )
    )
    # THEN the dna case id should have been mentioned in the logging (and used in the upload)

    # Assert that each string in 'expected_strings' is in 'caplog.text'
    for dna_case_id in dna_case_ids:
        assert (
            dna_case_id in caplog.text
        ), f"Expected '{dna_case_id}' to be in log output, but it wasn't."

    # THEN the api should know that it should find related DNA cases
    assert f"Finding DNA cases related to RNA case {rna_case_id}" in caplog.text

    # assert f"Running upload of report to DNA case {dna_case_ids[0]}" in caplog.text
    # THEN the report should be uploaded to Scout
    for case_id in dna_case_ids:
        assert (
            f"Uploading {ScoutCustomCaseReportTags.RNA_DELIVERY} report to Scout for case {case_id}"
            in caplog.text
        )


def test_upload_rna_report_to_not_yet_uploaded_dna_case_in_scout(
    caplog,
    rna_case_id: str,
    rna_store: Store,
    upload_mip_analysis_scout_api: UploadScoutAPI,
    mip_rna_analysis_hk_api: MockHousekeeperAPI,
):
    """Test that an error is raised when trying to upload an RNA report to a not yet uploaded DNA case."""

    caplog.set_level(logging.INFO)

    # GIVEN an RNA case, and an store with an RNA connected to it
    upload_mip_analysis_scout_api.status_db: Store = rna_store

    # GIVEN an RNA case with a multiqc-htlml report
    multiqc_file: File = mip_rna_analysis_hk_api.files(
        bundle=rna_case_id, tags=[ScoutCustomCaseReportTags.MULTIQC]
    )[0]

    # WHEN finding the related DNA case with no successful upload
    dna_case_ids: Set[str] = upload_mip_analysis_scout_api.get_unique_dna_cases_related_to_rna_case(
        case_id=rna_case_id
    )
    dna_case: Case = rna_store.get_case_by_internal_id(internal_id=list(dna_case_ids)[0])
    dna_case.analyses[0].uploaded_at = None

    # WHEN trying to upload the report
    # THEN a CgDataError should be raised
    with pytest.raises(CgDataError):
        upload_mip_analysis_scout_api.upload_rna_report_to_dna_case_in_scout(
            dry_run=False,
            rna_case_id=rna_case_id,
            report_type=ScoutCustomCaseReportTags.MULTIQC,
            report_file=multiqc_file,
        )



from pydantic.v1 import BaseModel

from cg.models.raredisease.raredisease import RarediseaseQCMetrics
from cg.models.rnafusion.rnafusion import RnafusionQCMetrics
from cg.models.taxprofiler.taxprofiler import TaxprofilerQCMetrics
from cg.models.tomte.tomte import TomteQCMetrics


class AnalysisModel(BaseModel):
    """Metadata analysis model"""

    class Config:
        arbitrary_types_allowed = True


class NextflowAnalysis(AnalysisModel):
    """Nextflow's analysis results model."""

    sample_metrics: dict[
        str, RarediseaseQCMetrics | RnafusionQCMetrics | TaxprofilerQCMetrics | TomteQCMetrics
    ]



import logging
from typing import Callable

from cg.constants.constants import SequencingQCStatus
from cg.services.sequencing_qc_service.utils import qc_bool_to_status
from typing import Callable

from cg.store.models import Case, Sample
from cg.services.sequencing_qc_service.quality_checks.checks import (
    run_quality_checks,
    get_sequencing_quality_check_for_case,
    get_sample_sequencing_quality_check,
)
from cg.store.store import Store

LOG = logging.getLogger(__name__)


class SequencingQCService:
    def __init__(self, store: Store):
        self.store = store

    def run_sequencing_qc(self) -> None:
        """Run QC for samples in pending or failed cases and store the aggregated score on each case."""
        cases: list[Case] = self.store.get_cases_for_sequencing_qc()
        for case in cases:
            passes_qc: bool = self.case_pass_sequencing_qc(case)
            qc_status: SequencingQCStatus = qc_bool_to_status(passes_qc)
            self.store.update_sequencing_qc_status(case=case, status=qc_status)
            LOG.info(f"Sequencing QC status for case {case.internal_id}: {qc_status}")

    @staticmethod
    def case_pass_sequencing_qc(case: Case) -> bool:
        """Run the sequencing QC for the case or sample."""
        sequencing_quality_check: Callable = get_sequencing_quality_check_for_case(case)
        return run_quality_checks(quality_checks=[sequencing_quality_check], case=case)

    @staticmethod
    def sample_pass_sequencing_qc(sample: Sample) -> bool:
        """Run the sequencing QC for a sample."""
        sample_sequencing_quality_check: Callable = get_sample_sequencing_quality_check()
        return run_quality_checks(quality_checks=[sample_sequencing_quality_check], sample=sample)



"""Delivery API context fixtures."""

from typing import Any

import pytest

from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.constants import DataDelivery, Workflow
from cg.models.cg_config import CGConfig
from cg.store.models import Case, Sample
from cg.store.store import Store
from tests.store_helpers import StoreHelpers


@pytest.fixture
def delivery_housekeeper_api(
    real_housekeeper_api: HousekeeperAPI,
    helpers: StoreHelpers,
    hk_delivery_sample_bundle: dict[str, Any],
    hk_delivery_another_sample_bundle: dict[str, Any],
    hk_delivery_case_bundle: dict[str, Any],
) -> HousekeeperAPI:
    """Delivery API Housekeeper context."""
    helpers.ensure_hk_bundle(
        store=real_housekeeper_api, bundle_data=hk_delivery_sample_bundle, include=True
    )
    helpers.ensure_hk_bundle(
        store=real_housekeeper_api, bundle_data=hk_delivery_another_sample_bundle, include=True
    )
    helpers.ensure_hk_bundle(
        store=real_housekeeper_api, bundle_data=hk_delivery_case_bundle, include=True
    )
    return real_housekeeper_api


@pytest.fixture
def delivery_store_balsamic(
    cg_context: CGConfig,
    helpers: StoreHelpers,
    no_sample_case_id: str,
    case_id: str,
    case_name: str,
    sample_id: str,
    another_sample_id: str,
    sample_name: str,
    another_sample_name: str,
    wgs_application_tag: str,
) -> Store:
    """Delivery API StatusDB context for Balsamic."""
    status_db: Store = cg_context.status_db

    # Error case without samples
    helpers.add_case(store=status_db, internal_id=no_sample_case_id, name=no_sample_case_id)

    # Balsamic case with FASTQ and analysis as data delivery
    case: Case = helpers.add_case(
        store=status_db,
        internal_id=case_id,
        name=case_name,
        data_analysis=Workflow.BALSAMIC,
        data_delivery=DataDelivery.FASTQ_ANALYSIS_SCOUT,
    )

    # Balsamic samples
    sample: Sample = helpers.add_sample(
        store=status_db,
        application_tag=wgs_application_tag,
        internal_id=sample_id,
        name=sample_name,
    )

    another_sample: Sample = helpers.add_sample(
        store=status_db,
        application_tag=wgs_application_tag,
        internal_id=another_sample_id,
        name=another_sample_name,
    )

    for sample_balsamic in [sample, another_sample]:
        helpers.add_relationship(store=status_db, case=case, sample=sample_balsamic)

    return status_db


@pytest.fixture
def delivery_store_microsalt(
    cg_context: CGConfig,
    helpers: StoreHelpers,
    case_id: str,
    no_sample_case_id: str,
    case_name: str,
    sample_id: str,
    another_sample_id: str,
    sample_id_not_enough_reads: str,
    total_sequenced_reads_pass: int,
    total_sequenced_reads_not_pass: int,
    sample_name: str,
    another_sample_name: str,
    microbial_application_tag: str,
) -> Store:
    """Delivery API StatusDB context for MicroSALT."""
    status_db: Store = cg_context.status_db

    # Error case without samples
    helpers.add_case(store=status_db, internal_id=no_sample_case_id, name=no_sample_case_id)

    # MicroSALT case with FASTQ-QC as data delivery
    case: Case = helpers.add_case(
        store=status_db,
        internal_id=case_id,
        name=case_name,
        data_analysis=Workflow.MICROSALT,
        data_delivery=DataDelivery.FASTQ_QC,
    )

    # MicroSALT samples
    sample: Sample = helpers.add_sample(
        store=status_db,
        application_tag=microbial_application_tag,
        internal_id=sample_id,
        name=sample_name,
        reads=total_sequenced_reads_pass,
    )

    another_sample: Sample = helpers.add_sample(
        store=status_db,
        application_tag=microbial_application_tag,
        internal_id=another_sample_id,
        name=another_sample_name,
        reads=total_sequenced_reads_pass,
    )

    sample_not_enough_reads: Sample = helpers.add_sample(
        store=status_db,
        application_tag=microbial_application_tag,
        internal_id=sample_id_not_enough_reads,
        reads=total_sequenced_reads_not_pass,
    )

    for sample_microsalt in [sample, another_sample, sample_not_enough_reads]:
        helpers.add_relationship(store=status_db, case=case, sample=sample_microsalt)

    return status_db


@pytest.fixture
def delivery_context_balsamic(
    cg_context: CGConfig,
    delivery_housekeeper_api: HousekeeperAPI,
    delivery_store_balsamic: Store,
) -> CGConfig:
    """Delivery API Balsamic context."""
    cg_context.housekeeper_api_ = delivery_housekeeper_api
    cg_context.status_db_ = delivery_store_balsamic
    return cg_context


@pytest.fixture
def delivery_context_microsalt(
    cg_context: CGConfig,
    delivery_housekeeper_api: HousekeeperAPI,
    delivery_store_microsalt: Store,
) -> CGConfig:
    """Delivery API MicroSALT context."""
    cg_context.housekeeper_api_ = delivery_housekeeper_api
    cg_context.status_db_ = delivery_store_microsalt
    return cg_context



"""Test for the time utils functions."""

import os
import time
from datetime import datetime
from pathlib import Path

import mock

from cg.constants.time import TWENTY_ONE_DAYS, TWENTY_ONE_DAYS_IN_SECONDS
from cg.utils.time import (
    is_directory_older_than_days_old,
    format_time_from_string,
    DATE_TIME_FORMAT,
    format_time_from_ctime,
)


def test_is_directory_older_than_days_old(tmp_path_factory):
    """Test whether a directory is older than a certain number of days."""
    # GIVEN a directory that is created now
    created_now_path: Path = tmp_path_factory.mktemp("created_now")

    # THEN checking whether a directory is older than 21 days returns true
    with mock.patch(
        "time.time",
        return_value=time.time() + TWENTY_ONE_DAYS_IN_SECONDS,
    ):
        assert is_directory_older_than_days_old(
            directory_path=created_now_path, days_old=TWENTY_ONE_DAYS
        )


def test_is_directory_not_older_than_days_old(tmp_path_factory):
    # GIVEN a directory that is created now
    created_now_path: Path = tmp_path_factory.mktemp("created_now")

    # THEN checking whether a directory is older than 21 days returns false
    assert not is_directory_older_than_days_old(
        directory_path=created_now_path, days_old=TWENTY_ONE_DAYS
    )


def test_format_time_from_string():
    """Test the format_time_from_string function."""
    # GIVEN a time stamp string
    time_stamp = "2021-01-01T12:00:00.000+00:00"

    # WHEN formatting the time stamp
    formatted_time = format_time_from_string(time_stamp)

    # THEN the time should be formatted correctly
    assert formatted_time == datetime(2021, 1, 1, 12, 0, 0)


def test_format_time_from_ctime(tmp_path_factory):
    """Test the format_time_from_ctime function."""
    # GIVEN a time stamp
    time_stamp: float = 1210521512.0

    # WHEN formatting the time stamp
    formatted_time = format_time_from_ctime(time_stamp)

    # THEN the time should be formatted correctly
    assert formatted_time == datetime.fromtimestamp(time_stamp)



from enum import Enum
from typing import Callable

from sqlalchemy.orm import Query

from cg.constants import PrepCategory, Workflow
from cg.store.models import Application


def filter_applications_by_tag(applications: Query, tag: str, **kwargs) -> Query:
    """Return application by tag."""
    return applications.filter(Application.tag == tag)


def filter_applications_is_not_archived(applications: Query, **kwargs) -> Query:
    """Return application which is not archived."""
    return applications.filter(Application.is_archived.is_(False))


def filter_applications_is_external(applications: Query, **kwargs) -> Query:
    """Return application which is external."""
    return applications.filter(Application.is_external == True)


def filter_applications_is_not_external(applications: Query, **kwargs) -> Query:
    """Return application which is not external."""
    return applications.filter(Application.is_external == False)


def filter_applications_has_versions(applications: Query, **kwargs) -> Query:
    return applications.filter(Application.versions.any())


def filter_application_by_prep_categories(
    applications: Query, prep_categories: list[PrepCategory], **kwargs
) -> Query:
    """Return application corresponding to dna sequencing."""
    return applications.filter(Application.prep_category.in_(prep_categories))


def apply_application_filter(
    filter_functions: list[Callable],
    applications: Query,
    tag: str = None,
    prep_category: str = None,
    prep_categories: list[Workflow] = None,
    entry_id: int = None,
) -> Query:
    """Apply filtering functions to the sample queries and return filtered results."""

    for filter_function in filter_functions:
        applications: Query = filter_function(
            applications=applications,
            tag=tag,
            prep_category=prep_category,
            prep_categories=prep_categories,
            entry_id=entry_id,
        )
    return applications


class ApplicationFilter(Enum):
    """Define Application filter functions."""

    HAS_VERSIONS = filter_applications_has_versions
    IS_EXTERNAL = filter_applications_is_external
    IS_NOT_EXTERNAL = filter_applications_is_not_external
    BY_TAG = filter_applications_by_tag
    IS_NOT_ARCHIVED = filter_applications_is_not_archived
    BY_PREP_CATEGORIES = filter_application_by_prep_categories



import logging
import zipfile
from pathlib import Path

LOG = logging.getLogger(__name__)


class Decompressor:

    def decompress(self, source_path: Path, destination_path: Path) -> Path:
        """Decompress a file to a directory."""
        if not source_path.exists():
            raise FileNotFoundError(f"Source path {source_path} does not exist.")
        with zipfile.ZipFile(source_path, "r") as zip_file:
            LOG.debug(f"Decompressing {source_path} to {destination_path}")
            zip_file.extractall(destination_path)
        if not self._decompressed_path_exists(destination_path):
            raise FileNotFoundError(
                f"Decompressed path {destination_path} does not exist, something went wrong."
            )
        return destination_path

    @staticmethod
    def _decompressed_path_exists(destination_path: Path) -> bool:
        """Check if the decompressed path already exists."""
        return destination_path.exists()



"""This file tests the analyses_to_delivery_report part of the status api"""

from cg.constants import DataDelivery, Workflow
from cg.constants.subject import PhenotypeStatus
from cg.store.models import CaseSample
from cg.store.store import Store
from cg.utils.date import get_date
from tests.store_helpers import StoreHelpers


def test_missing(analysis_store: Store, helpers: StoreHelpers, timestamp_now):
    """Test that analyses that are completed, but lacks a delivery report returned."""

    # GIVEN an analysis that is delivered but has no delivery report
    workflow = Workflow.BALSAMIC
    analysis = helpers.add_analysis(
        analysis_store,
        started_at=timestamp_now,
        completed_at=timestamp_now,
        uploaded_at=timestamp_now,
        workflow=workflow,
        data_delivery=DataDelivery.SCOUT,
    )
    sample = helpers.add_sample(analysis_store, delivered_at=timestamp_now)
    link: CaseSample = analysis_store.relate_sample(
        case=analysis.case, sample=sample, status=PhenotypeStatus.UNKNOWN
    )
    analysis_store.session.add(link)
    assert sample.delivered_at is not None
    assert analysis.delivery_report_created_at is None

    # WHEN calling the analyses_to_delivery_report
    analyses = analysis_store.analyses_to_delivery_report(workflow=workflow).all()

    # THEN this analyse should be returned
    assert analysis in analyses


def test_outdated_analysis(
    analysis_store: Store, helpers: StoreHelpers, timestamp_now, timestamp_yesterday
):
    """Tests that analyses that are older then when Hasta became production (2017-09-26) are not included in the cases to generate a delivery report for"""

    # GIVEN an analysis that is older than Hasta
    timestamp_old_analysis = get_date("2017-09-26")
    workflow = Workflow.BALSAMIC

    # GIVEN a delivery report created at date which is older than the upload date to trigger delivery report generation

    # GIVEN a store to add analysis to
    analysis = helpers.add_analysis(
        analysis_store,
        started_at=timestamp_old_analysis,
        uploaded_at=timestamp_now,
        delivery_reported_at=timestamp_yesterday,
        workflow=workflow,
    )

    # GIVEN samples which has been delivered
    sample = helpers.add_sample(analysis_store, delivered_at=timestamp_now)

    # GIVEN a store sample case relation
    link: CaseSample = analysis_store.relate_sample(
        case=analysis.case, sample=sample, status=PhenotypeStatus.UNKNOWN
    )
    analysis_store.session.add(link)

    # WHEN calling the analyses_to_delivery_report
    analyses = analysis_store.analyses_to_delivery_report(workflow=workflow).all()

    # THEN this analyses should not be returned
    assert len(analyses) == 0


def test_analyses_to_upload_delivery_reports(
    analysis_store: Store, helpers: StoreHelpers, timestamp_now
):
    """Tests extraction of analyses ready for delivery report upload"""

    # GIVEN an analysis that has a delivery report generated
    workflow = Workflow.BALSAMIC
    analysis = helpers.add_analysis(
        analysis_store,
        started_at=timestamp_now,
        completed_at=timestamp_now,
        uploaded_at=None,
        delivery_reported_at=timestamp_now,
        workflow=workflow,
        data_delivery=DataDelivery.FASTQ_ANALYSIS_SCOUT,
    )

    # WHEN calling the analyses_to_upload_delivery_reports
    analyses = analysis_store.analyses_to_upload_delivery_reports(workflow=workflow).all()

    # THEN the previously defined analysis should be returned
    assert analysis in analyses



import pytest

from cg.cli.post_process.utils import UnprocessedRunInfo
from cg.services.run_devices.pacbio.post_processing_service import PacBioPostProcessingService


@pytest.fixture
def pacbio_unprocessed_run_info(
    pacbio_sequencing_run_name: str, pac_bio_post_processing_service: PacBioPostProcessingService
) -> UnprocessedRunInfo:
    return UnprocessedRunInfo(
        name=pacbio_sequencing_run_name,
        post_processing_service=pac_bio_post_processing_service,
        instrument="pacbio",
    )


@pytest.fixture
def pacbio_barcoded_unprocessed_run_info(
    pacbio_barcoded_sequencing_run_name: str,
    pac_bio_post_processing_service: PacBioPostProcessingService,
) -> UnprocessedRunInfo:
    return UnprocessedRunInfo(
        name=pacbio_barcoded_sequencing_run_name,
        post_processing_service=pac_bio_post_processing_service,
        instrument="pacbio",
    )


@pytest.fixture
def pacbio_unprocessed_runs(
    pacbio_unprocessed_run_info: UnprocessedRunInfo,
    pacbio_barcoded_unprocessed_run_info: UnprocessedRunInfo,
) -> list[UnprocessedRunInfo]:
    return [pacbio_unprocessed_run_info, pacbio_barcoded_unprocessed_run_info]



from pathlib import Path

import pytest

from cg.constants import FileExtensions
from cg.utils.checksum.checksum import check_md5sum, extract_md5sum, is_md5sum_correct


def test_checksum(fastq_file: Path):
    """Tests if the function correctly calculates md5sum and returns the correct result."""
    # GIVEN a fastq file with a correct md5 file and a fastq file with an incorrect md5 file
    bad_md5sum_file_path: Path = fastq_file.parent.joinpath("fastq_run_R1_001.fastq.gz")

    # THEN a file with a correct md5 sum should return true
    assert check_md5sum(file_path=fastq_file, md5sum="a95cbb265540a2261fce941059784fd1")

    # THEN a file with an incorrect md5 sum should return false
    assert not check_md5sum(
        file_path=bad_md5sum_file_path, md5sum="c690b0124173772ec4cbbc43709d84ee"
    )


def test_extract_checksum(fastq_file: Path):
    """Tests if the function successfully extract the correct md5sum."""

    # GIVEN a file containing a md5sum
    md5sum_file = Path(f"{fastq_file.as_posix()}{FileExtensions.MD5}")

    # WHEN extracting the md5 sum
    extracted_sum: str = extract_md5sum(md5sum_file=md5sum_file)

    # THEN the function should return a string with the md5 sum
    assert isinstance(extracted_sum, str)


@pytest.mark.parametrize(
    "file_fixture, expected",
    [("fastq_file", True), ("fastq_file_father", False), ("non_existing_file_path", False)],
    ids=["Existing correct file", "Existing incorrect file", "Non existing file"],
)
def test_is_md5sum_correct(file_fixture: str, expected: bool, request: pytest.FixtureRequest):
    """Test that the function correctly checks if the md5sum is correct."""
    # GIVEN a file
    file_path: Path = request.getfixturevalue(file_fixture)

    # WHEN checking if the md5sum is correct
    is_correct: bool = is_md5sum_correct(file_path)

    # THEN a teh result is expected
    assert is_correct == expected



from enum import StrEnum

from cg.constants import ANALYSIS_SOURCES, METAGENOME_SOURCES
from cg.models.orders.order import OrderType

SEX_MAP = {"male": "M", "female": "F", "unknown": "unknown"}
REV_SEX_MAP = {value: key for key, value in SEX_MAP.items()}
CONTAINER_TYPES = ["Tube", "96 well plate"]
SOURCE_TYPES = set().union(METAGENOME_SOURCES, ANALYSIS_SOURCES)

CASE_PROJECT_TYPES = [
    OrderType.MIP_DNA,
    OrderType.BALSAMIC,
    OrderType.MIP_RNA,
]


class Orderform(StrEnum):
    BALSAMIC: str = "1508"
    BALSAMIC_QC: str = "1508"
    BALSAMIC_UMI: str = "1508"
    FASTQ: str = "1508"
    METAGENOME: str = "1508"
    FLUFFY: str = "1604"
    MICROSALT: str = "1603"
    MIP_DNA: str = "1508"
    MIP_RNA: str = "1508"
    RNAFUSION: str = "1508"
    RML: str = "1604"
    SARS_COV_2: str = "2184"
    MICROBIAL_FASTQ: str = "microbial_sequencing"

    @staticmethod
    def get_current_orderform_version(order_form: str) -> str:
        """Returns the current version of the given order form."""
        current_order_form_versions = {
            Orderform.MIP_DNA: "31",
            Orderform.RML: "18",
            Orderform.MICROSALT: "11",
            Orderform.SARS_COV_2: "9",
            Orderform.MICROBIAL_FASTQ: "1",
        }
        return current_order_form_versions[order_form]


REGION_CODES: dict[str, str] = {
    "Stockholm": "01",
    "Uppsala": "03",
    "Sörmland": "04",
    "Östergötland": "05",
    "Jönköpings län": "06",
    "Kronoberg": "07",
    "Kalmar län": "08",
    "Gotland": "09",
    "Blekinge": "10",
    "Skåne": "12",
    "Halland": "13",
    "Västra Götalandsregionen": "14",
    "Värmland": "17",
    "Örebro län": "18",
    "Västmanland": "19",
    "Dalarna": "20",
    "Gävleborg": "21",
    "Västernorrland": "22",
    "Jämtland Härjedalen": "23",
    "Västerbotten": "24",
    "Norrbotten": "25",
}

ORIGINAL_LAB_ADDRESSES: dict[str, str] = {
    "Unilabs Stockholm": "171 54 Solna",
    "Synlab Medilab": "183 53 Täby",
    "A05 Diagnostics": "171 65 Solna",
    "Karolinska University Hospital Solna": "171 76 Stockholm",
    "Karolinska University Hospital Huddinge": "141 86 Stockholm",
    "LaboratorieMedicinskt Centrum Gotland": "621 84 Visby",
    "Unilabs Eskilstuna Laboratorium": "631 88 Eskilstuna",
    "Norrland University Hospital": "901 85 Umeå",
}



import datetime as dt
import logging
import shutil
from pathlib import Path

import click
from dateutil.parser import parse as parse_date

from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.cli.utils import TOWER_WORKFLOW_TO_ANALYSIS_API_MAP
from cg.cli.workflow.utils import validate_force_store_option
from cg.constants import EXIT_FAIL, EXIT_SUCCESS, Workflow
from cg.constants.cli_options import COMMENT, DRY_RUN, FORCE, SKIP_CONFIRMATION
from cg.constants.observations import LOQUSDB_SUPPORTED_WORKFLOWS
from cg.exc import IlluminaRunsNeededError
from cg.meta.workflow.analysis import AnalysisAPI
from cg.meta.workflow.balsamic import BalsamicAnalysisAPI
from cg.meta.workflow.balsamic_pon import BalsamicPonAnalysisAPI
from cg.meta.workflow.balsamic_qc import BalsamicQCAnalysisAPI
from cg.meta.workflow.balsamic_umi import BalsamicUmiAnalysisAPI
from cg.meta.workflow.fluffy import FluffyAnalysisAPI
from cg.meta.workflow.microsalt import MicrosaltAnalysisAPI
from cg.meta.workflow.mip_dna import MipDNAAnalysisAPI
from cg.meta.workflow.mip_rna import MipRNAAnalysisAPI
from cg.meta.workflow.mutant import MutantAnalysisAPI
from cg.meta.workflow.nf_analysis import NfAnalysisAPI
from cg.meta.workflow.rnafusion import RnafusionAnalysisAPI
from cg.models.cg_config import CGConfig
from cg.services.deliver_files.rsync.service import (
    DeliveryRsyncService,
)
from cg.store.store import Store

ARGUMENT_BEFORE_STR = click.argument("before_str", type=str)
ARGUMENT_CASE_ID = click.argument("case_id", required=True)
ARGUMENT_WORKFLOW = click.argument("workflow", required=True)
OPTION_ANALYSIS_PARAMETERS_CONFIG = click.option(
    "--config-artic", type=str, help="Config with computational and lab related settings"
)
OPTION_LOQUSDB_SUPPORTED_WORKFLOW = click.option(
    "--workflow",
    type=click.Choice(LOQUSDB_SUPPORTED_WORKFLOWS),
    help="Limit observations upload to a specific workflow",
)

LOG = logging.getLogger(__name__)


@click.command("ensure-illumina-runs-on-disk")
@ARGUMENT_CASE_ID
@click.pass_obj
def ensure_illumina_runs_on_disk(context: CGConfig, case_id: str):
    """
    Check if Illumina runs are on disk for a given case.
    If not, request Illumina run and raise IlluminaRunsNeededError.
    """
    analysis_api: AnalysisAPI = context.meta_apis["analysis_api"]
    status_db: Store = context.status_db
    analysis_api.status_db.verify_case_exists(case_internal_id=case_id)
    if not status_db.are_all_illumina_runs_on_disk(case_id=case_id):
        if analysis_api.status_db.is_case_down_sampled(case_id=case_id):
            LOG.debug("All samples have been down sampled. Flow cell check not applicable")
            return
        elif analysis_api.status_db.is_case_external(case_id=case_id):
            LOG.debug("All samples are external. Flow cell check not applicable")
            return
        raise IlluminaRunsNeededError(
            "Analysis cannot be started: all flow cells need to be on disk to run the analysis"
        )
    LOG.info("All Illumina runs present on disk")


@click.command("resolve-compression")
@ARGUMENT_CASE_ID
@DRY_RUN
@click.pass_obj
def resolve_compression(context: CGConfig, case_id: str, dry_run: bool):
    """Handles cases where decompression is needed before starting analysis."""
    analysis_api: AnalysisAPI = context.meta_apis["analysis_api"]
    analysis_api.status_db.verify_case_exists(case_internal_id=case_id)

    analysis_api.resolve_decompression(case_id=case_id, dry_run=dry_run)


@click.command("link")
@ARGUMENT_CASE_ID
@DRY_RUN
@click.pass_obj
def link(context: CGConfig, case_id: str, dry_run: bool):
    """Link FASTQ files for all samples in a case."""
    analysis_api: AnalysisAPI = context.meta_apis["analysis_api"]
    analysis_api.status_db.verify_case_exists(case_internal_id=case_id)
    if dry_run:
        return
    analysis_api.link_fastq_files(case_id=case_id)


@click.command("store")
@ARGUMENT_CASE_ID
@COMMENT
@DRY_RUN
@FORCE
@click.pass_obj
def store(context: CGConfig, case_id: str, comment: str | None, dry_run: bool, force: bool):
    """Store finished analysis files in Housekeeper."""
    validate_force_store_option(force=force, comment=comment)
    analysis_api: AnalysisAPI = context.meta_apis["analysis_api"]
    housekeeper_api: HousekeeperAPI = context.housekeeper_api
    status_db: Store = context.status_db
    analysis_api.status_db.verify_case_exists(case_internal_id=case_id)

    if dry_run:
        LOG.info(f"Dry run: Would have stored deliverables for {case_id}")
        return
    try:
        analysis_api.upload_bundle_housekeeper(case_id=case_id, dry_run=dry_run, force=force)
        analysis_api.upload_bundle_statusdb(
            case_id=case_id, comment=comment, dry_run=dry_run, force=force
        )
        analysis_api.set_statusdb_action(case_id=case_id, action=None, dry_run=dry_run)
    except Exception as exception_object:
        housekeeper_api.rollback()
        status_db.session.rollback()
        LOG.error(f"Error storing deliverables for case {case_id} - {exception_object}")
        raise


@click.command("store-available")
@DRY_RUN
@click.pass_context
def store_available(context: click.Context, dry_run: bool) -> None:
    """Store bundles for all finished analyses in Housekeeper."""

    analysis_api: AnalysisAPI = context.obj.meta_apis["analysis_api"]

    exit_code: int = EXIT_SUCCESS
    for case_obj in analysis_api.get_cases_to_store():
        LOG.info(f"Storing deliverables for {case_obj.internal_id}")
        try:
            context.invoke(store, case_id=case_obj.internal_id, dry_run=dry_run)
        except Exception as exception_object:
            LOG.error(f"Error storing {case_obj.internal_id}: {exception_object}")
            exit_code = EXIT_FAIL
    if exit_code:
        raise click.Abort


@click.command("rsync-past-run-dirs")
@SKIP_CONFIRMATION
@DRY_RUN
@ARGUMENT_BEFORE_STR
@click.pass_obj
def rsync_past_run_dirs(
    context: CGConfig, before_str: str, dry_run: bool, skip_confirmation: bool
) -> None:
    """Remove deliver workflow commands."""

    rsync_api: DeliveryRsyncService = context.delivery_rsync_service

    before: dt.datetime = parse_date(before_str)

    for process in rsync_api.rsync_processes:
        if rsync_api.process_ready_to_clean(before=before, process=process):
            if skip_confirmation or click.confirm(f"Do you want to remove all files in {process}?"):
                if dry_run:
                    LOG.info(f"Would have removed {process}")
                    continue
                LOG.info(f"Removing {process.as_posix()}")
                shutil.rmtree(process, ignore_errors=True)
        else:
            LOG.info(f"{process.as_posix()} is still young")


@click.command("clean-run-dir")
@SKIP_CONFIRMATION
@DRY_RUN
@ARGUMENT_CASE_ID
@click.pass_obj
def clean_run_dir(context: CGConfig, skip_confirmation: bool, case_id: str, dry_run: bool = False):
    """Remove workflow run directory."""

    analysis_api: AnalysisAPI = context.meta_apis["analysis_api"]
    analysis_api.check_analysis_ongoing(case_id=case_id)

    analysis_path: list[Path] | Path = analysis_api.get_case_path(case_id)

    if dry_run:
        LOG.info(f"Would have deleted: {analysis_path}")
        return EXIT_SUCCESS

    analysis_api.clean_run_dir(
        case_id=case_id, skip_confirmation=skip_confirmation, case_path=analysis_path
    )


@click.command("past-run-dirs")
@SKIP_CONFIRMATION
@DRY_RUN
@ARGUMENT_BEFORE_STR
@click.pass_context
def past_run_dirs(
    context: click.Context, before_str: str, skip_confirmation: bool = False, dry_run: bool = False
):
    """Clean up of old case run dirs."""

    exit_code = EXIT_SUCCESS
    analysis_api: AnalysisAPI = context.obj.meta_apis["analysis_api"]

    before = parse_date(before_str)
    possible_cleanups = analysis_api.get_analyses_to_clean(before=before)
    LOG.info(f"Cleaning {len(possible_cleanups)} analyses created before {before}")

    for analysis in possible_cleanups:
        case_id = analysis.case.internal_id
        try:
            LOG.info(f"Cleaning {analysis_api.workflow} output for {case_id}")
            context.invoke(
                clean_run_dir, skip_confirmation=skip_confirmation, case_id=case_id, dry_run=dry_run
            )
        except FileNotFoundError:
            continue
        except Exception as error:
            LOG.error(f"Failed to clean directories for case {case_id} - {repr(error)}")
            exit_code = EXIT_FAIL

    if exit_code:
        raise click.Abort
    LOG.info(f"Done cleaning {analysis_api.workflow} output")


@click.command("balsamic-past-run-dirs")
@SKIP_CONFIRMATION
@DRY_RUN
@ARGUMENT_BEFORE_STR
@click.pass_context
def balsamic_past_run_dirs(
    context: click.Context, before_str: str, skip_confirmation: bool = False, dry_run: bool = False
):
    """Clean up of "old" Balsamic case run dirs."""

    context.obj.meta_apis["analysis_api"] = BalsamicAnalysisAPI(context.obj)
    context.invoke(
        past_run_dirs, skip_confirmation=skip_confirmation, dry_run=dry_run, before_str=before_str
    )


@click.command("balsamic-qc-past-run-dirs")
@SKIP_CONFIRMATION
@DRY_RUN
@ARGUMENT_BEFORE_STR
@click.pass_context
def balsamic_qc_past_run_dirs(
    context: click.Context, before_str: str, skip_confirmation: bool = False, dry_run: bool = False
):
    """Clean up of "old" Balsamic qc case run dirs."""

    context.obj.meta_apis["analysis_api"] = BalsamicQCAnalysisAPI(context.obj)
    context.invoke(
        past_run_dirs, skip_confirmation=skip_confirmation, dry_run=dry_run, before_str=before_str
    )


@click.command("balsamic-umi-past-run-dirs")
@SKIP_CONFIRMATION
@DRY_RUN
@ARGUMENT_BEFORE_STR
@click.pass_context
def balsamic_umi_past_run_dirs(
    context: click.Context, before_str: str, skip_confirmation: bool = False, dry_run: bool = False
):
    """Clean up of "old" Balsamic umi case run dirs."""

    context.obj.meta_apis["analysis_api"] = BalsamicUmiAnalysisAPI(context.obj)
    context.invoke(
        past_run_dirs, skip_confirmation=skip_confirmation, dry_run=dry_run, before_str=before_str
    )


@click.command("balsamic-pon-past-run-dirs")
@SKIP_CONFIRMATION
@DRY_RUN
@ARGUMENT_BEFORE_STR
@click.pass_context
def balsamic_pon_past_run_dirs(
    context: click.Context, before_str: str, skip_confirmation: bool = False, dry_run: bool = False
):
    """Clean up of "old" Balsamic pon case run dirs."""

    context.obj.meta_apis["analysis_api"] = BalsamicPonAnalysisAPI(context.obj)
    context.invoke(
        past_run_dirs, skip_confirmation=skip_confirmation, dry_run=dry_run, before_str=before_str
    )


@click.command("fluffy-past-run-dirs")
@SKIP_CONFIRMATION
@DRY_RUN
@ARGUMENT_BEFORE_STR
@click.pass_context
def fluffy_past_run_dirs(
    context: click.Context, before_str: str, skip_confirmation: bool = False, dry_run: bool = False
):
    """Clean up of "old" Fluffy case run dirs."""

    context.obj.meta_apis["analysis_api"] = FluffyAnalysisAPI(context.obj)
    context.invoke(
        past_run_dirs, skip_confirmation=skip_confirmation, dry_run=dry_run, before_str=before_str
    )


@click.command("mip-dna-past-run-dirs")
@SKIP_CONFIRMATION
@DRY_RUN
@ARGUMENT_BEFORE_STR
@click.pass_context
def mip_dna_past_run_dirs(
    context: click.Context, before_str: str, skip_confirmation: bool = False, dry_run: bool = False
):
    """Clean up of "old" MIP_DNA case run dirs."""

    context.obj.meta_apis["analysis_api"] = MipDNAAnalysisAPI(context.obj)
    context.invoke(
        past_run_dirs, skip_confirmation=skip_confirmation, dry_run=dry_run, before_str=before_str
    )


@click.command("mip-rna-past-run-dirs")
@SKIP_CONFIRMATION
@DRY_RUN
@ARGUMENT_BEFORE_STR
@click.pass_context
def mip_rna_past_run_dirs(
    context: click.Context, before_str: str, skip_confirmation: bool = False, dry_run: bool = False
):
    """Clean up of "old" MIP_RNA case run dirs."""

    context.obj.meta_apis["analysis_api"] = MipRNAAnalysisAPI(context.obj)
    context.invoke(
        past_run_dirs, skip_confirmation=skip_confirmation, dry_run=dry_run, before_str=before_str
    )


@click.command("mutant-past-run-dirs")
@SKIP_CONFIRMATION
@DRY_RUN
@ARGUMENT_BEFORE_STR
@click.pass_context
def mutant_past_run_dirs(
    context: click.Context, before_str: str, skip_confirmation: bool = False, dry_run: bool = False
):
    """Clean up of "old" MUTANT case run dirs."""

    context.obj.meta_apis["analysis_api"] = MutantAnalysisAPI(context.obj)
    context.invoke(
        past_run_dirs, skip_confirmation=skip_confirmation, dry_run=dry_run, before_str=before_str
    )


@click.command("rnafusion-past-run-dirs")
@SKIP_CONFIRMATION
@DRY_RUN
@ARGUMENT_BEFORE_STR
@click.pass_context
def rnafusion_past_run_dirs(
    context: click.Context, before_str: str, skip_confirmation: bool = False, dry_run: bool = False
):
    """Clean up of "old" RNAFUSION case run dirs."""

    context.obj.meta_apis["analysis_api"] = RnafusionAnalysisAPI(context.obj)
    context.invoke(
        past_run_dirs, skip_confirmation=skip_confirmation, dry_run=dry_run, before_str=before_str
    )


@click.command("microsalt-past-run-dirs")
@SKIP_CONFIRMATION
@DRY_RUN
@ARGUMENT_BEFORE_STR
@click.pass_context
def microsalt_past_run_dirs(
    context: click.Context, before_str: str, skip_confirmation: bool = False, dry_run: bool = False
):
    """Clean up of "old" microSALT case run dirs."""

    context.obj.meta_apis["analysis_api"]: MicrosaltAnalysisAPI = MicrosaltAnalysisAPI(context.obj)
    context.invoke(
        past_run_dirs, skip_confirmation=skip_confirmation, dry_run=dry_run, before_str=before_str
    )


@click.command("tower-past-run-dirs")
@SKIP_CONFIRMATION
@ARGUMENT_WORKFLOW
@ARGUMENT_BEFORE_STR
@click.pass_context
def tower_past_run_dirs(
    context: click.Context, before_str: str, workflow: Workflow, skip_confirmation: bool = False
):
    """Clean up of "old" tower case run dirs."""
    if workflow not in TOWER_WORKFLOW_TO_ANALYSIS_API_MAP:
        LOG.error(f"Please ensure that the provided workflow {workflow} is using Tower")
        raise click.Abort()
    analysis_api: NfAnalysisAPI = TOWER_WORKFLOW_TO_ANALYSIS_API_MAP.get(workflow)(context.obj)
    analysis_api.clean_past_run_dirs(before_date=before_str, skip_confirmation=skip_confirmation)
